"""
Local update rule
"""

import heapq
from .fodge_utils import *


def lur_embed(current_node, proj_nodes, dict_proj, dim, dict_enode_enode, G, epsilon):
    """
    Calculate the current node embedding according to the formuala mentioned in the pdf file
    :param current_node: Calculated embedding node
    :param proj_nodes: Nodes that are currently in the embedding
    :param dict_proj: Dictionary of nodes and their embeddings
    :param dim: Embedding Dimension
    :param dict_enode_enode: key==(node that is in the projection and has neighbors in it), value==(set of neighbors
    that are in projection)
    :param G: The graph (networkx)
    :param epsilon: The weight that is given to the second order neighbours.
    :return: The node's final embedding
    """
    proj = []
    weights_first_neigh = []
    mean_two_order_proj = []
    # get a list of the projections of the neighbors in the projection
    proj_nodes1 = list(proj_nodes)
    # the number of first order neighbors
    k1 = len(proj_nodes)
    k2 = 0
    # to calculate the mean projection of the second order neighbors
    for k in range(len(proj_nodes1)):
        if dict_enode_enode.get(proj_nodes1[k]) is not None:
            two_order_neighs = dict_enode_enode.get(proj_nodes1[k]).copy()
            if current_node in two_order_neighs:
                two_order_neighs.remove(current_node)
            if len(two_order_neighs) == 0:
                two_order_neighs = None
        else:
            two_order_neighs = None
        # if the neighbors in the projection also have neighbors in the projection calculate the average projection
        if two_order_neighs is not None:
            two_order_neighs = list(two_order_neighs)
            k2 += len(two_order_neighs)
            two_order_projs = []
            weights_two_neigh = []
            for i in range(len(two_order_neighs)):
                two_order_proj = dict_proj[two_order_neighs[i]]
                weights_two_neigh.append(G[proj_nodes1[k]][two_order_neighs[i]]["weight"])
                two_order_projs.append(two_order_proj)
            two_order_projs = np.array(two_order_projs)
            weights_two_neigh = np.array(weights_two_neigh)
            two_order_projs = np.average(two_order_projs, axis=0, weights=weights_two_neigh)
        # else, the mean projection in 0
        else:
            two_order_projs = np.zeros(dim)
        mean_two_order_proj.append(two_order_projs)
        # list of embeddings of each first order neighbour
        proj.append(dict_proj[proj_nodes1[k]])
        weights_first_neigh.append(G[current_node][proj_nodes1[k]]["weight"])
    # for every neighbor we have the average projection of its neighbors, so now do average on all of them
    mean_two_order_proj = np.array(mean_two_order_proj)
    mean_two_order_proj = np.mean(mean_two_order_proj, axis=0)
    proj = np.array(proj)
    weights_first_neigh = np.array(weights_first_neigh)
    # find the mean proj of first order neighbours
    proj = np.average(proj, axis=0, weights=weights_first_neigh)
    # the final projection of the node
    final_proj = proj + epsilon * (k2 / k1) * (proj - mean_two_order_proj)
    return final_proj


def one_iteration(dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, dict_cc, dict_nodes_cc,
                  set_n_e, current_node, dim, G, epsilon):
    """
    A function that does one iteration over a given batch
    """
    condition = 1

    # check cc
    list_cc, dict_nodes_cc, dict_cc = update_dicts_cc(current_node, dict_node_enode, dict_cc, dict_nodes_cc)

    # get the neighbors in projection of node i
    embd_neigh = dict_node_enode[current_node]
    # the final projection of the node
    final_proj = lur_embed(current_node, embd_neigh, dict_enode_proj, dim, dict_enode_enode, G, epsilon)
    # add the node and its projection to the dict of projections
    dict_enode_proj.update({current_node: final_proj})
    # add our node to the dict of proj to proj and delete it from node_enode because now it's in the projection
    dict_enode_enode.update({current_node: embd_neigh})
    dict_node_enode.pop(current_node)
    # get the non embd neighbors of the node
    relevant_n_e = dict_node_node[current_node]
    # delete because now it is in the projection
    dict_node_node.pop(current_node)
    embd_neigh = list(embd_neigh)
    for i in range(len(embd_neigh)):
        f = dict_enode_enode.get(embd_neigh[i])
        if f is not None:
            dict_enode_enode[embd_neigh[i]].update([current_node])
        else:
            dict_enode_enode.update({embd_neigh[i]: set([current_node])})
    # check if num of non embd neighbors of our node bigger then zero
    if len(relevant_n_e) > 0:
        # loop of non embd neighbors
        relevant_n_e1 = list(relevant_n_e)
        for j in range(len(relevant_n_e)):
            tmp_append_n_n = dict_node_node.get(relevant_n_e1[j])
            if tmp_append_n_n is not None:
                # if relevant_n_e1[j] in dict_node_node:
                tmp_append_n_n = tmp_append_n_n - set([current_node])
                dict_node_node[relevant_n_e1[j]] = tmp_append_n_n
            tmp_append = dict_node_enode.get(relevant_n_e1[j])
            if tmp_append is not None:
                # add our node to the set cause now our node is in embd
                tmp_append.update(set([current_node]))
                dict_node_enode[relevant_n_e1[j]] = tmp_append
            else:
                dict_node_enode.update({relevant_n_e1[j]: set([current_node])})
    set_n_e.remove(current_node)

    return condition, dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, set_n_e, list_cc, dict_cc, \
           dict_nodes_cc


def one_iteration_for_connected_cc(dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, set_n_e,
                                   current_node, dim, G, epsilon):
    """
    A function that does one iteration over a given batch for a specific new connected component (for merging)
    """
    condition = 1
    # get the neighbors in projection of node i
    embd_neigh = dict_node_enode[current_node]
    # the final projection of the node
    final_proj = lur_embed(current_node, embd_neigh, dict_enode_proj, dim, dict_enode_enode, G, epsilon)
    # add the node and its projection to the dict of projections
    dict_enode_proj.update({current_node: final_proj})
    # add our node to the dict of proj to proj and delete it from node_enode because now it's in the projection
    dict_enode_enode.update({current_node: embd_neigh})
    dict_node_enode.pop(current_node)
    # get the non embd neighbors of the node
    relevant_n_e = dict_node_node[current_node]
    # delete because now it is in the projection
    dict_node_node.pop(current_node)
    embd_neigh = list(embd_neigh)
    for i in range(len(embd_neigh)):
        f = dict_enode_enode.get(embd_neigh[i])
        if f is not None:
            dict_enode_enode[embd_neigh[i]].update([current_node])
        else:
            dict_enode_enode.update({embd_neigh[i]: set([current_node])})
    # check if num of non embd neighbors of our node bigger then zero
    if len(relevant_n_e) > 0:
        # loop of non embd neighbors
        relevant_n_e1 = list(relevant_n_e)
        for j in range(len(relevant_n_e)):
            tmp_append_n_n = dict_node_node.get(relevant_n_e1[j])
            if tmp_append_n_n is not None:
                # if relevant_n_e1[j] in dict_node_node:
                tmp_append_n_n = tmp_append_n_n - set([current_node])
                dict_node_node[relevant_n_e1[j]] = tmp_append_n_n
            tmp_append = dict_node_enode.get(relevant_n_e1[j])
            if tmp_append is not None:
                # add our node to the set cause now our node is in embd
                tmp_append.update(set([current_node]))
                dict_node_enode[relevant_n_e1[j]] = tmp_append
            else:
                dict_node_enode.update({relevant_n_e1[j]: set([current_node])})
    set_n_e.remove(current_node)
    return condition, dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, set_n_e


def final_function_lur(dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, dict_cc, dict_nodes_cc,
                       set_n_e, batch_precent, dim, G, epsilon):
    """
    the final function that iteratively divided the dictionary of nodes without embedding into number of batches
    determined by batch_precent. It does by building a heap every iteration so that we enter the nodes to the
    projection from the nodes which have the most neighbors in the embedding to the least. This way the projection
    gets more accurate.
    """
    condition = 1
    k = 0
    set_n_e2 = set_n_e.copy()
    while condition > 0:
        condition = 0
        k += 1
        batch_size = int(batch_precent * len(set_n_e2))
        # loop over node are not in the embedding
        if batch_size > len(set_n_e):
            num_times = len(set_n_e)
        else:
            num_times = batch_size
        list_n_e = list(set_n_e)
        heap = []
        for i in range(len(list_n_e)):
            my_node = list_n_e[i]
            a = dict_node_enode.get(my_node)
            if a is not None:
                num_neighbors = len(dict_node_enode[my_node])
            else:
                num_neighbors = 0
            heapq.heappush(heap, [-num_neighbors, my_node])
        for i in range(len(set_n_e))[:num_times]:
            # look on node number i in the loop
            current_node = heapq.heappop(heap)[1]
            if dict_node_enode.get(current_node) is not None:
                condition, dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, set_n_e, list_cc, \
                dict_cc, dict_nodes_cc = one_iteration(dict_enode_proj, dict_node_enode, dict_node_node,
                                                       dict_enode_enode, dict_cc, dict_nodes_cc, set_n_e, current_node,
                                                       dim, G, epsilon)
                # if list_cc > 0 then the node connects between several connected components and they should be
                # embed all over again considering the new connections
                if len(list_cc) > 0:
                    list_all_nodes = []
                    for j in range(len(list_cc) - 1):
                        for pw in dict_cc[list_cc[j + 1]]:
                            list_all_nodes.append(pw)
                    set_no_proj_nodes = set(list_all_nodes)
                    dict_enode_proj, set_no_e = final_function_lur_for_connected_cc(dict_enode_proj, dict_node_enode,
                                                                                     dict_node_node,
                                                                                     dict_enode_enode,
                                                                                     set_no_proj_nodes, batch_precent,
                                                                                     dim, G, epsilon)
                    for n in list_all_nodes:
                        dict_nodes_cc[n] = list_cc[0]
                    for j in range(len(list_cc) - 1):
                        for p in dict_cc[list_cc[j + 1]]:
                            dict_cc[list_cc[0]].append(p)
                        if dict_cc.get(list_cc[j + 1]) is not None:
                            del dict_cc[list_cc[j + 1]]
    return dict_enode_proj, dict_nodes_cc, dict_cc, set_n_e


def final_function_lur_for_connected_cc(dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, set_n_e,
                                        batch_precent, dim, G, epsilon):
    """
    the final function that iteratively divided the dictionary of nodes without embedding into number of batches
    determined by batch_precent. It does by building a heap every iteration so that we enter the nodes to the
    projection from the nodes which have the most neighbors in the embedding to the least. This way the projection
    gets more accurate.
    """
    condition = 1
    k = 0
    set_n_e2 = set_n_e.copy()
    while condition > 0:
        condition = 0
        k += 1
        batch_size = int(batch_precent * len(set_n_e2))
        # loop over node are not in the embedding
        if batch_size > len(set_n_e):
            num_times = len(set_n_e)
        else:
            num_times = batch_size
        list_n_e = list(set_n_e)
        heap = []
        for i in range(len(list_n_e)):
            my_node = list_n_e[i]
            a = dict_node_enode.get(my_node)
            if a is not None:
                num_neighbors = len(dict_node_enode[my_node])
            else:
                num_neighbors = 0
            heapq.heappush(heap, [-num_neighbors, my_node])
        for i in range(len(set_n_e))[:num_times]:
            # look on node number i in the loop
            current_node = heapq.heappop(heap)[1]
            if dict_node_enode.get(current_node) is not None:
                condition, dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode, set_n_e = \
                    one_iteration_for_connected_cc(dict_enode_proj, dict_node_enode, dict_node_node, dict_enode_enode,
                                                   set_n_e, current_node, dim, G, epsilon)
    return dict_enode_proj, set_n_e


"""
Main file of FODGE
"""

from .LUR import *
from .fodge_utils import *
import time
from .load_data import *
import sys


class FODGE:
    """
    Class to run one of our suggested dynamic embedding methods.
    """

    def __init__(self, name, graph_path, save_path, func=data_loader, initial_method="node2vec", dim=128, epsilon=0.01,
                 alpha_exist=0., beta=0.99, number=0, mission=None, file_tags=None):
        """
        Init function to initialize the class.
        :param name: Name of the dataset (string).
        :param graph_path: Path to where the dataset file is.
        :param save_path: Path to where to save the calculated embedding
        :param func: Loader function to load the dataset as a dictionary of snapshots. In order to create your own
                     load function, one should define it in the file load_data.py. (Name of the function)
        :param initial_method: Initial state-of-the-art algorithm to embed the first snapshot with. Options are
                               "node2vec", "HOPE", "GAE", "GF" and "GCN".
        :param dim: Embedding dimension (int)
        :param epsilon: The weight that is given to the second order neighbours.
        :param alpha_exist: Weight that is given to the previous changed existing nodes embeddings when they are
        recalculated (float between 0 and 1).
        :param beta: Rate of exponential decay of the edges weights through time
        :param number: How many vertices in the cumulative initial snapshot (choose a number where a 5-core exists)
        :param mission: None if it is not for a specific mission, "lp" for Temporal Link Prediction task and "nc" for
                        node classification task.
        :param file_tags: If GCN GEA is used, then one should profile a file of tags
        """
        self.name = name
        # create dict of snapshots containing the edges to each time and corresponding dict of weights
        self.dict_snapshots, self.dict_weights = self.load_graph(func, (name, graph_path))

        if file_tags is not None:
            self.graph = self.create_weighted_graph_for_all_times()
            nodes = list(self.graph.nodes())
            self.nc_nodes = []
            self.nodes_no_nc = list(set(nodes) - set(self.nc_nodes))

        self.initial_method = initial_method
        self.file_tags = file_tags
        self.number = number
        self.beta = beta
        self.epsilon = epsilon
        self.dim = dim
        self.alpha_exist = alpha_exist
        self.params_dict = self.define_params_for_initial_method()

        self.g_list, self.nodes_list, T, self.index = self.create_cumulative_graphs(number)

        if self.index > 0:
            self.change_dict_snapshots(T)

        print("The temporal network after creating the cumulative initial snapshot:")
        keys = list(self.dict_snapshots.keys())
        for i in range(len(keys)):
            print(i, keys[i])

        print("Ready to embed the {} temporal network with FODGE".format(name))

        if mission is None:
            # calculate the graph embedding and return a dictionary of nodes as keys and embedding vectors as values,
            self.full_dict_embeddings, self.dict_all_embeddings, self.total_time = self.calculate_embedding()
            self.save_embedding(save_path)

    @staticmethod
    def load_graph(func, t):
        """
        Function to load the graph as a dictionary of snapshots
        """
        dict_snapshots, dict_weights = func(*t)
        return dict_snapshots, dict_weights

    def define_params_for_initial_method(self):
        """
        According to the initial state-of-the-art embedding method, create the dictionary of parameters.
        :return: Parameters dictionary
        """
        if self.initial_method == "node2vec":
            params_dict = {"dimension": self.dim, "walk_length": 80, "num_walks": 16, "workers": 2}
        elif self.initial_method == "GF":
            params_dict = {"dimension": self.dim, "eta": 0.1, "regularization": 0.1, "max_iter": 1000,
                           "print_step": 100}
        elif self.initial_method == "HOPE":
            params_dict = {"dimension": self.dim, "beta": 0.1}
        elif self.initial_method == "GAE":
            params_dict = {"model": "gcn_vae", "seed": 42, "epochs": 70, "hidden1": 256, "hidden2": self.dim, "lr": 0.1,
                           "dropout": 0.2, "dataset_str": self.name, "dimension": self.dim}
        elif self.initial_method == "GCN":
            params_dict = {"dimension": self.dim, "epochs": 100, "lr": 0.03, "weight_decay": 0, "dropout": 0.2,
                           "nc_nodes": self.nc_nodes}
        else:
            sys.exit("Chosen GEA is not valid. Please choose one of the following: node2vec, HOPE, GF, GEA. If you have"
                     "tags to your data, you can also choose GCN")
        return params_dict

    def change_weights(self, t1, t2, H):
        """
        Change weights between consecutive snapshots according to exponential decay rate
        """
        edges1 = set(self.dict_snapshots[t1])
        edges2 = set(self.dict_snapshots[t2])
        common_edges = edges1.intersection(edges2)
        new_edges = list(edges2 - common_edges)
        old_edges = list(edges1 - common_edges)
        for e in list(common_edges):
            H[e[0]][e[1]]["weight"] = 1
            H[e[1]][e[0]]["weight"] = 1
        for n in new_edges:
            H[n[0]][n[1]]["weight"] = 1
            H[n[1]][n[0]]["weight"] = 1
        for o in old_edges:
            w = H[o[0]][o[1]]["weight"]
            H[o[0]][o[1]]["weight"] = self.beta * w
            H[o[1]][o[0]]["weight"] = self.beta * w
        return H

    def create_cumulative_graphs(self, number):
        """
        Create the temporal network gamma=[G1,G2,...,GT]
        """
        index = 0
        T = None
        g_list = []
        nodes_list = []
        times = list(self.dict_snapshots.keys())
        for i in range(len(times)):
            if i == 0:
                G = nx.DiGraph()
                G.add_edges_from(self.dict_snapshots[times[0]])
                weights = self.dict_weights[times[i]]
                G = add_weights(G, weights)
                H = G.to_undirected()
                g_list.append(H.copy())
                nodes = list(H.copy().nodes())
                nodes_list.append(nodes)
            else:
                H.add_edges_from(self.dict_snapshots[times[i]])
                H.to_undirected()
                H = self.change_weights(times[i - 1], times[i], H)
                H_copy = H.copy()
                g_list.append(H_copy)
                nodes_list.append(list(H_copy.nodes()))
                if H_copy.number_of_nodes() < number:
                    index = i
                    T = H_copy.copy()
            print(i, g_list[i].number_of_nodes(), g_list[i].number_of_edges())
        if number != 0:
            g_list = g_list[index:]
            nodes_list = nodes_list[index:]

        return g_list, nodes_list, T, index

    def change_dict_snapshots(self, H):
        """
        If the first snapshot is a union of a few graphs, delete these times and put the only the first time with all
        the edges of the union graph.
        """
        edges = list(H.edges())
        times_to_delete = list(self.dict_snapshots.keys())[:self.index]
        for t in times_to_delete:
            del self.dict_snapshots[t]
        self.dict_snapshots[list(self.dict_snapshots.keys())[0]] = edges

    def create_graph_for_all_times(self):
        """
        Create a "static" graph with all nodes and edges from all snapshots.
        """
        edges = []
        times = list(self.dict_snapshots.keys())
        for t in times:
            for edge in self.dict_snapshots[t]:
                edges.append(edge)
        G = nx.DiGraph()
        G.add_edges_from(edges)
        H = G.to_undirected()
        return H

    def create_weighted_graph_for_all_times(self):
        """
        Create a "static" graph with all nodes and edges from all snapshots. (weighted)
        """
        edges = []
        weights = {}
        times = list(self.dict_snapshots.keys())
        for t in times:
            for edge in self.dict_snapshots[t]:
                if weights.get(edge) is None:
                    weights.update({edge: 1})
                    weights.update({(edge[1], edge[0]): 1})
                else:
                    weights[edge] += 1
                    weights[(edge[1], edge[0])] += 1
                edges.append(edge)
        G = nx.DiGraph()
        G.add_edges_from(edges)
        H = G.to_undirected()
        edges = list(H.edges())
        for e in edges:
            H[e[0]][e[1]]["weight"] = weights[e]
            H[e[1]][e[0]]["weight"] = weights[e]
        return H

    def calculate_embedding(self):
        """
        Function to calculate the embedding of the given graph and return:
        - full_dict_embeddings - A dictionary of all nodes embedding from all snapshots. If a node shows up in more
          than one snapshot so its final embedding in this dictionary is its embedding in the last time stamp the node
          has been shown,
        - dict_all_embeddings: A dictionary of dictionary. The keys are times and the value for each key is the
          embedding dictionary for the specific time.
        - total_time: Running time in seconds for the whole embedding calculation of our suggested method.
        """
        full_dict_embeddings, dict_all_embeddings, total_time = \
            main_fodge(self.initial_method, self.dict_snapshots, self.g_list, self.nodes_list, self.dim,
                       self.params_dict, epsilon=self.epsilon, alpha_exist=self.alpha_exist, file_tags=self.file_tags)
        return full_dict_embeddings, dict_all_embeddings, total_time

    def save_embedding(self, path, mission=None):
        """
        Save the calculated embedding in a .npy file.
        1) full_dict_embeddings: Dict of the embeddings of all the nodes at the final time.
        2) dict_all_embeddings: Dict of the embeddings at any given time. Keys are times and values are dicts of the
        embeddings of all the nodes that appear specifically at this time.
        :param path: Path to where to save the embedding
        :param mission: lp for temporal link prediction task, else None
        :return:
        """
        m = mission if mission is not None else ""
        curr_time = datetime.now().strftime("%Y%m%d_%H%M")
        file_name = self.name + " " + m + " + " + self.initial_method + " +n=" + str(self.number) + " + a=" + str(
            self.alpha_exist) + " + b=" + str(self.beta) + " + e=" + str(self.epsilon) + " " + str(curr_time)
        np.save(os.path.join(path, '{}.npy'.format(file_name)), self.full_dict_embeddings)
        np.save(os.path.join(path, '{}_all_embeddings.npy'.format(file_name)), self.dict_all_embeddings)


def main_fodge(initial_method, dict_snapshots, g_list, nodes_list, dim, params, epsilon=0.01, alpha_exist=0.,
               file_tags=None):
    """
    Main function to run the our suggested dynamic embedding method - FODGE. The inputs variables were already explained
    in the initialization function. This function returns 2 embedding dicts and running time in seconds as explained in
    the "save_embedding" method of the previous class.
    """
    total_time = 0
    t = time.time()

    # start with the initial cumulative snapshot embedding
    dict_projections = initial_function(initial_method, g_list, nodes_list, params, file_tags=file_tags)
    print("number of nodes in projection: ", len(dict_projections))

    # create dictionaries of connected components
    dict_cc, dict_nodes_cc = create_dict_cc_nodes_cc(dict_projections)

    # calculate neighbours dict for each snapshot
    list_neighbours_dict = []
    for i in range(len(g_list)):
        list_neighbours_dict.append(create_dict_neighbors(g_list[i]))

    counter = 1

    # First initialization of the wanted embedding dictionaries and total running time
    times, dict_all_embeddings, full_dict_embeddings = create_two_embedding_dicts(dict_snapshots, dict_projections)

    set_proj_nodes = set(nodes_list[0])

    highest_k_core_all = [get_initial_proj_nodes_by_k_core(g_list[i], 100) for i in range(len(g_list))]

    for i in range(len(g_list) - 1):
        print("Snapshot number", counter)
        # save the dict of embeddings
        if i > 0:
            dict_projections = z.copy()

        j_s = jaccard_similarity(highest_k_core_all[i], highest_k_core_all[i + 1])
        print("js is: ", j_s)

        # perform rotation if needed
        if j_s < 0.9:
            sub_g = g_list[i + 1].subgraph(highest_k_core_all[i + 1])
            print("number of nodes", sub_g.number_of_nodes(), "number of edges", sub_g.number_of_edges())
            print("project core")
            _, dict_proj_core_i_1, _ = final(sub_g, initial_method, params, file_tags=file_tags)
            dict_proj_core_i = dict_of_core(highest_k_core_all[i], dict_projections)

            new_dict_proj_core = rotation(dict_proj_core_i_1, dict_proj_core_i)

            for h in highest_k_core_all[i + 1]:
                dict_projections.update({h: new_dict_proj_core[h]})
                full_dict_embeddings.update({h: new_dict_proj_core[h]})

        # separate nodes of each snapshot to new, exist and disappear (according to the next snapshot)
        new, exist, disappear = divide_snapshot_nodes(nodes_list[i], nodes_list[i + 1])

        # neighbours dictionary for each snapshot
        neighbors_dict = list_neighbours_dict[i + 1]

        # existing nodes that their connections are changes need to be treated as new ones
        changed_exist, changed_exist_proj_dict, dict_projections = check_changed_existing(exist,
                                                                                          list_neighbours_dict[i],
                                                                                          neighbors_dict,
                                                                                          dict_projections)

        set_nodes_no_proj = set(new).union(set(changed_exist))
        set_proj_nodes = set_proj_nodes - set(disappear) - set(changed_exist)
        list_proj_nodes = list(set_proj_nodes)

        # each time update the embedding dict according to the current time stamp
        new_dict = {l: dict_projections[l] for l in list_proj_nodes}
        dict_projections = new_dict.copy()

        # create dicts of connections
        dict_node_node, dict_node_enode, dict_enode_enode = create_dicts_of_connections \
            (set_proj_nodes, set_nodes_no_proj, neighbors_dict)

        # calculate each new node embedding
        dict_projections, dict_nodes_cc, dict_cc, set_no_proj = \
            final_function_lur(dict_projections, dict_node_enode, dict_node_node, dict_enode_enode, dict_cc,
                               dict_nodes_cc, set_nodes_no_proj, 1, dim, g_list[i + 1], epsilon)

        # update which nodes are in the embedding now
        set_proj_nodes = set(nodes_list[i + 1])

        # only for the existing nodes (i.e. nodes that are in sequential numbers) update their embedding according
        # to their last and new embedding (only if neighbours were changed)
        dict_projections = changes_existing_nodes(changed_exist, alpha_exist, dict_projections,
                                                  changed_exist_proj_dict)

        # if there are nodes that are not in the final embedding it means there is a new connected component that
        # needs to be embed from scratch with state-of-the-art embedding algorithm
        if len(set_no_proj) > 1:
            dict_cc, dict_nodes_cc, z = create_new_connected_component(dict_projections, dict_cc, dict_nodes_cc,
                                                                       g_list, set_no_proj, initial_method, params,
                                                                       i, file_tags=file_tags)
        else:
            z = dict_projections.copy()

        # to calculate running time
        elapsed_time = time.time() - t
        total_time += elapsed_time

        counter += 1

        # update wanted dictionaries
        dict_all_embeddings.update({times[counter-1]: z.copy()})
        full_dict_embeddings = update_full_dict_projections(list(z.keys()), full_dict_embeddings, z)

        # print useful information
        print("running time: ", elapsed_time)
        print("The number of nodes that aren't in the final projection:", len(set_no_proj))
        print("The number of nodes that are in the final projection:", len(z), "\n")

    return full_dict_embeddings, dict_all_embeddings, total_time



"""
Utils file for first temporal link prediction task
"""

import random
import numpy as np
import sys
import csv
import networkx as nx


def nodes_test_in_train(nodes_list, index):
    """
    Given list of nodes for each time stamp and an index separating between train and test examples, return the nodes
    in the test set that are also in the training set.
    :param nodes_list: List of lists of nodes for each time stamp.
    :param index: Index indicating the time stamp that afterwards it's the test set, and beforehand it's the train set.
    :return: A list of nodes in the test set that are also in the train set
    """
    nodes_train = []
    nodes_test = []

    for i in range(len(nodes_list)):
        if i < index:
            for j in nodes_list[i]:
                nodes_train.append(j)
        else:
            for j in nodes_list[i]:
                nodes_test.append(j)

    nodes_train = set(dict.fromkeys(nodes_train))
    nodes_test = set(dict.fromkeys(nodes_test))
    nodes = nodes_train.intersection(nodes_test)

    return nodes


def create_full_lists_missing_edges(index, non_edges_file, nodes, start_index, dict_snapshots, dict_embeddings):
    """
    Create lists of missing edges for both train and test.
    :param index: Index of pivot time- until pivot time (including) it is train set, afterwards it is test set.
    :param nodes: List of nodes of the graph
    :param non_edges_file: Csv file containing list of non edges (explained in the initialization of the class).
    :param start_index: Index of the cumulative initial snapshot
    :param dict_embeddings: Dict of embeddings, key is a vertex id and value is its numpy array embedding
    :param dict_snapshots: Dict of snapshots, key is a time and value is a list of edges occurring at this time
    :return: List of missing edges for both train and test sets.
    """
    missing_edges_train = calculate_list_of_false_edges(index, "train", nodes, start_index, dict_snapshots,
                                                        dict_embeddings, non_edges_file=non_edges_file)
    missing_edges_test = calculate_list_of_false_edges(index, "test", nodes, start_index, dict_snapshots,
                                                       dict_embeddings, non_edges_file=non_edges_file)
    return missing_edges_train, missing_edges_test


def choose_false_edges(non_edges, K):
    """
    From the list of non-edges choose K false edges.
    :param non_edges: List of non edges (train or test, depends on the input).
    :param K: Number of non-edges to choose randomly.
    :return List of K non-edges.
    """
    indexes = random.sample(range(0, len(non_edges)), K)
    false_edges = []
    for i in indexes:
        false_edges.append(non_edges[i])
    return false_edges


def create_data_for_link_prediction(K_train, K_test, missing_edges_train, missing_edges_test):
    """
    Create false train and test edges for link prediction task.
    :param K_train: Number of true/false edges to choose from the train set, is equal to the total number of edges
                    in the train set (for all time stamps).
    :param K_test: Number of true/false edges to choose from the test set, is equal to the total number of edges
                   in the test set (for all time stamps).
    :param missing_edges_train: List of missing edges in the train set
    :param missing_edges_test: List of missing edges in the test set
    :param non_edges_file: Csv file containing list of non edges (explained in the initialization of the class).
    :return: List of K non-edges chosen randomly from the whole lists of non edges, both for train and test.
    """
    false_edges_train = choose_false_edges(missing_edges_train, K_train)
    false_edges_test = choose_false_edges(missing_edges_test + missing_edges_train, K_test)
    return false_edges_train, false_edges_test


def remove_true_false_edges(dict_snapshots, dict_weights, index):
    """
    Remove chosen true edges from the graph so the embedding could be calculated without them.
    :param dict_snapshots: Dict where keys are times and values are a list of edges for each time stamp.
    :param dict_weights: Dict where keys are times and values are list of weights for each edge in the time stamp, order
                         corresponds to the order of edges in dict_snapshots.
    :param index: Index of pivot time- until pivot time (including) it is train set, afterwards it is test set.
    :return: Updated dict_snapshots and dict_weights.
    """
    times = list(dict_snapshots.keys())
    mapping = {i: times[i] for i in range(len(times))}
    keys = list(mapping.keys())
    for key in keys:
        if key < index:
            continue
        else:
            del dict_snapshots[mapping[key]]
            del dict_weights[mapping[key]]
    return dict_snapshots, dict_weights


def for_lp_mission(name, graph, nodes_list, dict_snapshots, dict_weights, r, start_index):
    """
    Create initial data for the link prediction task- index pivot time, nodes common to train and test set,
    dict of edges for each snapshot without the chosen test positive edges, number of edges in train and test.
    :param name: Name of the dataset
    :param graph: Our graph
    :param nodes_list: List of all nodes in the graph (for all time stamps)
    :param dict_snapshots: Dict where keys are times and values are a list of edges for each time stamp.
    :param dict_weights: Weights of the edges at each time
    :param r: Wanted test ratio (r = K_test / (K_train + K_test)
    :param start_index: The index of the cumulative graph
    :return: Explained above.
    """
    # this is a function that finds the index that times before are train and times after is test, but it is better
    # determine ahead because it takes a long time to run it
    try:
        index = choose_test_index(name) - start_index
    except:
        pivot_time, index, mapping = get_pivot_time(graph, dict_snapshots, wanted_ratio=r, min_ratio=0.1)
    print("pay attention the index is !!!!!!!!", index)
    nodes = nodes_test_in_train(nodes_list, index)
    print("len nodes is ", len(nodes))
    K_train, K_test, true_edges_train, true_edges_test = train_test_edges(dict_snapshots, index, nodes)
    dict_snapshots, dict_weights = remove_true_false_edges(dict_snapshots, dict_weights, index)
    return index, nodes, dict_snapshots, true_edges_train, true_edges_test, K_train, K_test, dict_weights


def calculate_list_of_false_edges(index, type_, nodes, start_index, dict_snapshots, dict_projections,
                                  non_edges_file=None):
    """
    Get a list of non edges.
    :param index: Representing the time stamp that separates between nodes in train and nodes in test
    :param type_: "train" - for train nodes, "test"- for test nodes
    :param nodes: Set of nodes common to test and train set.
    :param non_edges_file: CSV file where each row is time,source,target. To create such file please see
    "calculate_non_edges.py" file.
    :param start_index: The index of the cumulative graph
    :param dict_snapshots: Dict where keys are times and values are a list of edges for each time stamp.
    :param dict_projections: Dict of embeddings, key is a vertex id and value is its numpy array embedding
    :return: A list of non-edges
    """
    l = len(dict_snapshots.keys()) + start_index
    mapping = {}
    for j in range(l):
        if j < start_index:
            mapping.update({j: 0})
        else:
            mapping.update({j: j - start_index})

    if non_edges_file is None:
        sys.exit("you have to create non_edges_file")
    else:
        non_edges = []
        print("Starting taking non_edges, may take a while (5-10 minutes), stay put!")
        csvfile = open(non_edges_file, 'r', newline='')
        obj = csv.reader(csvfile)
        for row in obj:
            if mapping.get(int(row[0])) is not None:
                if type_ == "train":
                    if mapping[int(row[0])] < index:
                        if dict_projections.get(row[1]) is not None and dict_projections.get(row[2]) is not None:
                            if len(set([row[1], row[2]]).intersection(nodes)) == 2:
                                non_edges.append((mapping[int(row[0])], (row[1], row[2])))
                else:
                    if mapping[int(row[0])] >= index:
                        if dict_projections.get(row[1]) is not None and dict_projections.get(row[2]) is not None:
                            if len(set([row[1], row[2]]).intersection(nodes)) == 2:
                                non_edges.append((mapping[int(row[0])], (row[1], row[2])))
    print("example for non edges: ", non_edges[0])
    return non_edges


def create_x_y(dict_proj, true_edges, false_edges, K, d):
    """
    Create X, Y for linear regression
    :param dict_proj: Dict of embeddings (vertex: embedding)
    :param true_edges: List of true edges
    :param false_edges: List of false edges
    :param K: Number of train/test edges
    :param d: Embedding dimension
    :return: X, Y for linear regression
    """
    nodes = list(dict_proj.keys())
    dict_edges = {n: [] for n in nodes}
    X = np.zeros(shape=(2 * K, 2 * d))
    Y = np.zeros(shape=(2 * K, 1))
    count = 0
    for m in true_edges:
        edge = m[1]
        embd1 = dict_proj[edge[0]]
        embd2 = dict_proj[edge[1]]
        con_embed = np.concatenate((embd1, embd2), axis=0)
        X[count, :] = con_embed
        Y[count, 0] = int(1)
        count += 1
        dict_edges[edge[0]].append(count)
        dict_edges[edge[1]].append(count)
    for m in false_edges:
        edge = m[1]
        embd1 = dict_proj[edge[0]]
        embd2 = dict_proj[edge[1]]
        con_embed = np.concatenate((embd1, embd2), axis=0)
        X[count, :] = con_embed
        Y[count, 0] = int(0)
        count += 1
        dict_edges[edge[0]].append(count)
        dict_edges[edge[1]].append(count)
    return X, Y, dict_edges


def create_mapping(dict_times):
    """
    If times are not integers transform each one of them to a unique integer of its own.
    """
    keys = list(dict_times.keys())
    mapping = {}
    for i in range(len(keys)):
        mapping.update({keys[i]: i})
    return mapping


def get_graph_T(graph, dict_edges, min_time=-np.inf, max_time=np.inf, mapping=None):
    """
    Given 2 times, return the graph that has all nodes and edges that have been in the snapshots in between (including
    the maximum time).
    :param graph: The given graph
    :param dict_edges: Dict where keys==times and values==list of edges for each time stamp.
    :param min_time: Minimum time
    :param max_time: Maximum time
    :param mapping: If times are not integers, it is a mapping dictionary between the name of the times and indexes.
    :return: The new graph between the two times, and the new dict_times between these 2 times.
    """
    relevant_edges = []
    relevant_edges_no_time = []
    new_dict_edges = {}

    if len(graph.nodes()) == 0:
        return graph

    keys = list(dict_edges.keys())

    for key in keys:
        t = key
        if mapping is not None:
            t = mapping[t]
        if min_time < t <= max_time:
            for edge in dict_edges[key]:
                u = edge[0]
                v = edge[1]
                relevant_edges.append((u, v, t))
                relevant_edges_no_time.append((u, v))
                if new_dict_edges.get(t) is None:
                    new_dict_edges.update({t: [(u, v)]})
                else:
                    new_dict_edges[t].append((u, v))
        else:
            continue

    new_graph = nx.DiGraph()
    new_graph.add_edges_from(relevant_edges_no_time)

    return new_graph, new_dict_edges


def get_graph_times(dict_time_edge):
    """
    Get all time stamps.
    """
    keys = list(dict_time_edge.keys())
    # keys.sort()
    return keys


def get_pivot_time(graph, dict_edges, wanted_ratio=0.2, min_ratio=0.1):
    """
    Given a graph and dictionary of snapshots (keys are times and values are the graph's edges in that time), calculate
    the pivot time that gives a wanted ratio to the train and test edges.
    :param graph: Our graph
    :param dict_edges: Dict where keys==times and values==list of edges for each time stamp.
    :param wanted_ratio: Wanted test ratio (r = K_test / (K_train + K_test)
    :param min_ratio: If we can't find time for the wanted ratio, try for the minimum ratio.
    :return: The pivot time and its index.
    """
    times = get_graph_times(dict_edges)
    mapping = {}
    for i in range(len(times)):
        mapping.update({times[i]: i})

    if wanted_ratio == 0:
        return times[-1]

    time2dist_from_ratio = {}
    for time in times[int(len(times) / 3):]:
        if mapping is not None:
            time = mapping[time]
        train_graph, _ = get_graph_T(graph, dict_edges, max_time=time, mapping=mapping)
        num_edges_train = train_graph.number_of_edges()

        test_graph, _ = get_graph_T(graph, dict_edges, min_time=time, mapping=mapping)
        test_graph.remove_nodes_from(
            [node for node in list(test_graph.nodes()) if node not in list(train_graph.nodes())])
        num_edges_test = test_graph.number_of_edges()

        current_ratio = num_edges_test / (num_edges_test + num_edges_train)
        print(2)

        if current_ratio <= min_ratio:
            continue

        time2dist_from_ratio.update({time: np.abs(wanted_ratio - current_ratio)})

    pivot_time = min(time2dist_from_ratio, key=time2dist_from_ratio.get)

    if mapping is not None:
        for key, value in mapping.items():
            if value == pivot_time:
                true_time = key
    else:
        true_time = 0

    print(f"pivot time {pivot_time}, is close to the wanted ratio by {round(time2dist_from_ratio[pivot_time], 3)}")

    return true_time, pivot_time, mapping


def train_test_edges(dict_edges_by_time, index_pivot_time, nodes):
    """
    Function to return a list of false edges and true edges, each list has tuples in the form: (time, (source, target))
    :param dict_edges_by_time: Dict where keys==times and values==list of edges for each time stamp.
    :param index_pivot_time: Representing the time stamp that separates between nodes in train and nodes in test
    :param nodes: Nodes appear both in the train and test sets.
    """
    times = get_graph_times(dict_edges_by_time)
    train_times = times[:index_pivot_time]
    test_times = times[index_pivot_time:]

    train_dict_times = {}
    test_dict_times = {}

    for t in train_times:
        train_dict_times.update({t: dict_edges_by_time[t]})
    for t in test_times:
        test_dict_times.update({t: dict_edges_by_time[t]})

    l1 = 0
    train_edges = []
    keys = list(train_dict_times.keys())
    for j in range(len(keys)):
        key = keys[j]
        for i in train_dict_times[key]:
            if len(set([i[0], i[1]]).intersection(nodes)) == 2:
                l1 += 1
                train_edges.append((j, i))

    l2 = 0
    test_edges = []
    keys = list(test_dict_times.keys())
    for j in range(len(keys)):
        key = keys[j]
        for s in range(len(test_dict_times[key])):
            i = test_dict_times[key][s]
            if len(set([i[0], i[1]]).intersection(nodes)) == 2:
                l2 += 1
                test_edges.append((j + index_pivot_time, i))

    return l1, l2, train_edges, test_edges


def choose_test_index(name):
    if name == "facebook-wosn-wall":
        val = 42
    elif name == "facebook_friendships":
        val = 23
    elif name == "ia-enron-email-dynamic":
        val = 43
    elif name == "dblp":
        val = 24
    elif name == "wiki-talk-temporal":
        val = 72
    elif name == "fb-wosn-friends":
        val = 24
    elif name == "sx-mathoverflow":
        val = 60
    return val


"""
Utils file for first temporal link prediction task to calculate the scores.
"""

import numpy as np


def calculate_avg_score(score, rounds):
    """
    Given the lists of scores for every round of every split, calculate the average score of every split.
    :param score: F1-micro / F1-macro / Accuracy / Auc
    :param rounds: How many times the experiment has been applied for each split.
    :return: Average score for every split
    """
    all_avg_scores = []
    for i in range(score.shape[1]):
        avg_score = (np.sum(score[:, i])) / rounds
        all_avg_scores.append(avg_score)
    return all_avg_scores


def calculate_all_avg_scores_lp(micro, macro, acc, auc, rounds):
    """
    For all scores calculate the average score for every split. The function returns list for every
    score type- 1 for cheap node2vec and 2 for regular node2vec.
    """
    all_avg_micro = calculate_avg_score(micro, rounds)
    all_avg_macro = calculate_avg_score(macro, rounds)
    all_avg_acc = calculate_avg_score(acc, rounds)
    all_avg_auc = calculate_avg_score(auc, rounds)
    return all_avg_micro, all_avg_macro, all_avg_acc, all_avg_auc


def initialize_scores():
    """
    Helper function to initialize the scores for link prediction mission
    """
    my_micro = [0, 0, 0, 0, 0]
    my_macro = [0, 0, 0, 0, 0]
    my_acc = [0, 0, 0, 0, 0]
    my_auc = [0, 0, 0, 0, 0]
    return my_micro, my_macro, my_acc, my_auc


def first_help_calculate_lp(score, avg_score):
    """
    Helper function for scores calculation
    """
    score = [x + y for x, y in zip(score, avg_score)]
    return score


def second_help_calculate_lp(score, number_of_sub_graphs):
    """
    Helper function for scores calculation
    """
    score = [x / number_of_sub_graphs for x in score]
    return score


import numpy as np
import scipy.sparse as sp
import torch
import scipy
import networkx as nx
from scipy import io


def encode_onehot(labels):
    classes = set(labels)
    classes_dict = {c: np.identity(len(classes))[i, :] for i, c in
                    enumerate(classes)}
    labels_onehot = np.array(list(map(classes_dict.get, labels)),
                             dtype=np.int32)
    return labels_onehot


def read_yelp_labels(file_tags):
    """
    Read labels of yelp dataset
    """
    X = np.loadtxt(file_tags)
    Y = np.int_(X)
    number_of_labels = Y.shape[1]
    for k in range(number_of_labels):
        if np.all((Y[:, k] == 0), axis=0):
            Y = np.delete(Y, k, 1)
    return Y
    

def read_flickr_labels(file_tags):
    features_struct = scipy.io.loadmat(file_tags)
    labels = scipy.sparse.csr_matrix(features_struct["group"])
    a = scipy.sparse.csr_matrix.toarray(labels)
    return a


def read_labels_nodes(file_tags):
    X = np.genfromtxt(file_tags, dtype=np.dtype(str))
    d_X = {"labels": X[:, -1], "nodes": X[:, 0]}
    return d_X


def new_load_data(G, sub_G, file_tags, num_of_nodes, nc_nodes, multilabel=False, features_file=None):
    nodes = list(sub_G.nodes())
    n2i = {nodes[i]: i for i in range(len(nodes))}
    common = list(set(nodes).intersection(set(nc_nodes)))
    train_nodes = list(set(nodes)-set(common))
    
    all_labels = {}
    with open(file_tags, 'r') as reader:
        for line in reader:
            parts = line.strip().split()
            n_id, l_id = str(parts[0]), str(parts[1])
            all_labels[n_id] = l_id
    reader.close()
    labels_d = {}
    for node in nodes:
        labels_d.update({node: all_labels[node]})
    labels = list(labels_d.values())
    labels = np.asarray(labels, dtype=np.dtype(str))
    labels = encode_onehot(labels)
        
    if features_file is None:
        features = sp.identity(num_of_nodes, dtype=np.float32)
    else:
        idx_features_labels = np.genfromtxt("../datasets/cora.content", dtype=np.dtype(str))
        features = sp.csr_matrix(idx_features_labels[:, 1:-1], dtype=np.float32)
    
    idx_map = {j: i for i, j in enumerate(nodes)}
    adj = sp.coo_matrix(nx.to_scipy_sparse_matrix(sub_G))

    # build symmetric adjacency matrix
    adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)

    features = normalize(features)
    adj = normalize(adj + sp.eye(adj.shape[0]))
    
    idx_train = [n2i[t] for t in train_nodes]
    idx_val = [n2i[t] for t in common]
    idx_test = [n2i[t] for t in common]

    features = torch.FloatTensor(np.array(features.todense()))
    if multilabel:
        labels = torch.FloatTensor(labels)
    else:
        labels = torch.LongTensor(np.where(labels)[1])
    
    adj = sparse_mx_to_torch_sparse_tensor(adj)

    idx_train = torch.LongTensor(idx_train)
    idx_val = torch.LongTensor(idx_val)
    idx_test = torch.LongTensor(idx_test)

    return adj, features, labels, idx_train, idx_val, idx_test


def normalize(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    return mx


def accuracy(output, labels):
    preds = output.max(1)[1].type_as(labels)
    correct = preds.eq(labels).double()
    correct = correct.sum()
    return correct / len(labels)


def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)



"""
Main function to run FODGE.
In the terminal run
    python main.py -h
This will print a message with explanations on all relevant parameters you should insert and how to write the command
to run this file
"""

import argparse
from .evaluation_tasks.temporal_link_prediction import *
from .evaluation_tasks.calculate_non_edges import *
from .evaluation_tasks.linear_regression_link_prediction import *


parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--name', default='facebook_friendships', type=str, help='Name of the dataset (str)')
parser.add_argument('--datasets_path', default='datasets', type=str, help='Path to where the dataset file is (str)')
parser.add_argument('--save_path', default='embeddings', type=str,
                    help='Path to where to save the calculated embedding (str)')
parser.add_argument('--initial_method', default='node2vec', type=str,
                    help='Initial GEA to embed the first snapshot with. Options are node2vec, HOPE, GAE, GF. If the'
                         'graph has tags, GCN option can also be used (str)')
parser.add_argument('--dim', default=128, type=int, help='Embedding dimension (int)')
parser.add_argument('--epsilon', default=0.01, type=float,
                    help='Relative weight given to first and second neighbors in the local update rule (float)')
parser.add_argument('--alpha', default=0.2, type=float,
                    help='The weight given to the recent embedding when calculating the current one (float)')
parser.add_argument('--beta', default=0.3, type=float, help='The rate of the exponential decay of the weights (float)')
parser.add_argument('--number', default=1000, type=int,
                    help='How many vertices in the cumulative initial snapshot (choose a number where a 5-core exists)'
                         '(int)')
parser.add_argument('--file_tags', default=None, type=str,
                    help='If GCN GEA is used, then one should provide the path of the file of tags (str)')
parser.add_argument('--link_prediction_1', default="False", type=str,
                    help='True if you want to perform temporal link prediction task (type 1 with neural network), else '
                         'False')
parser.add_argument('--link_prediction_2', default="False", type=str,
                    help='True if you want to perform temporal link prediction task (type 2 with linear regression),'
                         'else False')
parser.add_argument('--test_ratio', default=0.2, type=float, help='Test ratio for temporal link prediction tasks '
                                                                  '(relevant to both) both)')
parser.add_argument('--val_ratio', default=0.3, type=float, help='Val ratio for temporal link prediction task '
                                                                 '(relevant only to type 2- with linear regression)')
parser.add_argument('--non_edges_file', default='non_edges_facebook_friendships.csv', type=str,
                    help='Name of non edges csv file as explained in the readme. If you do not have any, insert None '
                         '(str) and it will be created during the running (can take a while). relevant only to type 1- '
                         'with linear regression')


args = parser.parse_args()

func = data_loader

if args.link_prediction_1 == "False" and args.link_prediction_2 == "False":
    # Only embed the temporal network
    DE = FODGE(args.name, args.datasets_path, args.save_path, func=func, initial_method=args.initial_method, dim=args.dim,
               epsilon=args.epsilon, alpha_exist=args.alpha, beta=args.beta, number=args.number,
               file_tags=args.file_tags)
elif args.link_prediction_1 == "True":
    # Perform first temporal link prediction
    if args.non_edges_file == "None":  # create non_edges_file if the user does not give one
        non_edges_file = create_non_edges_file(args.name, args.datasets_path, func)
        path_non_edges_file = os.path.join("evaluation_tasks", non_edges_file)
    else:
        path_non_edges_file = os.path.join("evaluation_tasks", args.non_edges_file)
    LP = TemporalLinkPrediction1(args.name, args.datasets_path, args.save_path, func=func,
                                 initial_method=args.initial_method, dim=args.dim, epsilon=args.epsilon,
                                 alpha_exist=args.alpha, beta=args.beta, number=args.number, test_ratio=args.test_ratio,
                                 non_edges_file=path_non_edges_file, file_tags=args.file_tags)
elif args.link_prediction_2 == "True":
    LP = TemporalLinkPrediction2(args.name, args.save_path, args.datasets_path, func=func,
                                 initial_method=args.initial_method, dim=args.dim, epsilon=args.epsilon,
                                 alpha_exist=args.alpha, beta=args.beta, number=args.number, test_ratio=args.test_ratio,
                                 val_ratio=args.val_ratio, file_tags=args.file_tags)




import torch
import torch.nn.modules.loss
import torch.nn.functional as F


def loss_function(preds, labels, mu, logvar, n_nodes, norm, pos_weight):
    pos_weight = torch.tensor(pos_weight)
    cost = norm * F.binary_cross_entropy_with_logits(preds, labels, pos_weight=pos_weight)

    # see Appendix B from VAE paper:
    # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
    # https://arxiv.org/abs/1312.6114
    # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
    KLD = -0.5 / n_nodes * torch.mean(torch.sum(
        1 + 2 * logvar - mu.pow(2) - logvar.exp().pow(2), 1))
    return cost + KLD



from __future__ import division
from __future__ import print_function
import my_init
import argparse
import time

import numpy as np
import scipy.sparse as sp
import torch
from torch import optim

from gae.model import GCNModelVAE
from gae.optimizer import loss_function
from gae.utils import load_data, mask_test_edges, preprocess_graph, get_roc_score

parser = argparse.ArgumentParser()
parser.add_argument('--model', type=str, default='gcn_vae', help="models used")
parser.add_argument('--seed', type=int, default=42, help='Random seed.')
parser.add_argument('--epochs', type=int, default=200, help='Number of epochs to train.')
parser.add_argument('--hidden1', type=int, default=32, help='Number of units in hidden layer 1.')
parser.add_argument('--hidden2', type=int, default=16, help='Number of units in hidden layer 2.')
parser.add_argument('--lr', type=float, default=0.01, help='Initial learning rate.')
parser.add_argument('--dropout', type=float, default=0., help='Dropout rate (1 - keep probability).')
parser.add_argument('--dataset-str', type=str, default='cora', help='type of dataset.')

args = parser.parse_args()


def gae_for(args):
    adj, features = load_data(args.dataset_str)
    n_nodes, feat_dim = features.shape

    # Store original adjacency matrix (without diagonal entries) for later
    adj_orig = adj
    adj_orig = adj_orig - sp.dia_matrix((adj_orig.diagonal()[np.newaxis, :], [0]), shape=adj_orig.shape)
    adj_orig.eliminate_zeros()

    adj_train, train_edges, val_edges, val_edges_false, test_edges, test_edges_false = mask_test_edges(adj)
    adj = adj_train

    # Some preprocessing
    adj_norm = preprocess_graph(adj)
    adj_label = adj_train + sp.eye(adj_train.shape[0])
    # adj_label = sparse_to_tuple(adj_label)
    adj_label = torch.FloatTensor(adj_label.toarray())

    pos_weight = float(adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum()
    norm = adj.shape[0] * adj.shape[0] / float((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)

    model = GCNModelVAE(feat_dim, args.hidden1, args.hidden2, args.dropout)
    optimizer = optim.Adam(model.parameters(), lr=args.lr)

    hidden_emb = None
    for epoch in range(args.epochs):
        t = time.time()
        model.train()
        optimizer.zero_grad()
        recovered, mu, logvar, z = model(features, adj_norm)
        loss = loss_function(preds=recovered, labels=adj_label,
                             mu=mu, logvar=logvar, n_nodes=n_nodes,
                             norm=norm, pos_weight=pos_weight)
        loss.backward()
        cur_loss = loss.item()
        optimizer.step()

        hidden_emb = mu.data.numpy()
        roc_curr, ap_curr = get_roc_score(hidden_emb, adj_orig, val_edges, val_edges_false)

        print("Epoch:", '%04d' % (epoch + 1), "train_loss=", "{:.5f}".format(cur_loss),
              "val_ap=", "{:.5f}".format(ap_curr),
              "time=", "{:.5f}".format(time.time() - t)
              )
    
    embedding = z.detach().numpy()
    print("Optimization Finished!")

    roc_score, ap_score = get_roc_score(hidden_emb, adj_orig, test_edges, test_edges_false)
    print('Test ROC score: ' + str(roc_score))
    print('Test AP score: ' + str(ap_score))


if __name__ == '__main__':
    gae_for(args)



"""
Implantation of 5 state-of-the-art static embedding algorithms: Node2Vec, Graph Factorization, HOPE, GAE and GCN.
- The implementations of GF and HOPE were taken from GEM toolkit (https://github.com/palash1992/GEM)
- Node2vec is implemented using node2vec public package (https://github.com/eliorc/node2vec)
- The implementation of GCN was taken from their public github repository (https://github.com/tkipf/pygcn)
- The implementation of GAE was taken from their public github repository (https://github.com/tkipf/gae)
"""
from __future__ import division
from __future__ import print_function
import sys
from time import time
from node2vec import Node2Vec
from torch import optim
from scipy.sparse import identity
from scipy.sparse.linalg import inv
from scipy.sparse.linalg import svds
from sklearn.metrics import accuracy_score
from .gae_pytorch.gae.model import GCNModelVAE
from .gae_pytorch.gae.optimizer import loss_function
from .gae_pytorch.gae.utils import load_data, mask_test_edges, preprocess_graph, get_roc_score, create_false_edges
from .gcn.models import *
from .gcn.utils import *



class GAE:
    def __init__(self, params, method_name, graph):
        self._d = params["hidden1"]
        self._method_name = method_name
        self._graph = graph
        self.create_adj_features()
        self._params = params
        self.embedding_matrix, self.embedding_dict = self.train()

    def create_adj_features(self):
        self._adj = nx.to_scipy_sparse_matrix(self._graph)
        self._features = torch.tensor(np.identity(self._graph.number_of_nodes())).type('torch.FloatTensor')

    def train(self):
        adj, features = self._adj, self._features
        n_nodes, feat_dim = features.shape

        nodes = list(self._graph.nodes())
        if len(nodes) > 100:
            hidden1 = self._params['hidden1']
        else:
            hidden1 = 100

        # Store original adjacency matrix (without diagonal entries) for later
        adj_orig = adj
        adj_orig = adj_orig - sp.dia_matrix((adj_orig.diagonal()[np.newaxis, :], [0]), shape=adj_orig.shape)
        adj_orig.eliminate_zeros()

        # adj_train, train_edges, val_edges, val_edges_false, test_edges, test_edges_false = mask_test_edges(adj)
        adj_train, train_edges, train_edges_false = create_false_edges(adj)
        adj = adj_train

        # Some preprocessing
        adj_norm = preprocess_graph(adj)
        adj_label = adj_train + sp.eye(adj_train.shape[0])
        # adj_label = sparse_to_tuple(adj_label)
        adj_label = torch.FloatTensor(adj_label.toarray())

        pos_weight = float(adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum()
        norm = adj.shape[0] * adj.shape[0] / float((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)

        model = GCNModelVAE(feat_dim, hidden1, self._params['hidden2'], self._params['dropout'])
        optimizer = optim.Adam(model.parameters(), lr=self._params['lr'])

        hidden_emb = None
        for epoch in range(self._params['epochs']):
            t = time()
            model.train()
            optimizer.zero_grad()
            recovered, mu, logvar, z = model(features, adj_norm)
            loss = loss_function(preds=recovered, labels=adj_label,
                                 mu=mu, logvar=logvar, n_nodes=n_nodes,
                                 norm=norm, pos_weight=pos_weight)
            loss.backward()
            cur_loss = loss.item()
            optimizer.step()

            hidden_emb = mu.data.numpy()
            # roc_curr, ap_curr = get_roc_score(hidden_emb, adj_orig, val_edges, val_edges_false)
            roc_curr, ap_curr = get_roc_score(hidden_emb, adj_orig, train_edges, train_edges_false)

        embedding_matrix = z.detach().numpy()
        embedding_dict = {nodes[i]: embedding_matrix[i, :] for i in range(self._graph.number_of_nodes())}

        return embedding_matrix, embedding_dict


def train_(epoch, num_of_epochs, model, optimizer, loss, features, labels, adj, idx_train, idx_val, multilabel=False):
    """
    Train function for GCN.
    """
    t = time()
    optimizer.zero_grad()
    h, output = model(features, adj)
    if not multilabel:
        loss_train = loss(output[idx_train], labels[idx_train])
        acc_train = accuracy(output[idx_train], labels[idx_train])
    else:
        loss_train = 0
        acc_train = 0
        for j in range(labels.shape[0]):
            loss_train += loss(output[j, :], labels[j, :])
        loss_train /= labels.shape[0]
        output = output.detach().numpy()
        labels = labels.detach().numpy()
        for i in range(output.shape[0]):
            for j in range(output.shape[1]):
                if output[i, j] > 0.5:
                    output[i, j] = 1
                else:
                    output[i, j] = 0
            acc_train += accuracy_score(labels[i], output[i])
        acc_train /= output.shape[0]
    loss_train.backward()
    optimizer.step()

    print('Epoch: {:04d}'.format(epoch + 1),
          'loss_train: {:.4f}'.format(loss_train.item()),
          'acc_train: {:.4f}'.format(acc_train.item()), 'time: {:.4f}s'.format(time() - t))

    if epoch == num_of_epochs - 1:
        return h  # layer before last layer
        # return x  # layer before softmax


class StaticGraphEmbedding:
    def __init__(self, d, method_name, graph):
        """
        Initialize the Embedding class
        :param d: dimension of embedding
        """
        self._d = d
        self._method_name = method_name
        self._graph = graph

    @staticmethod
    def get_method_name(self):
        """
        Returns the name for the embedding method
        :param self:
        :return: The name of embedding
        """
        return self._method_name

    def learn_embedding(self):
        """
        Learning the graph embedding from the adjacency matrix.
        :param graph: the graph to embed in networkx DiGraph format
        :return:
        """
        pass

    @staticmethod
    def get_embedding(self):
        """
        Returns the learnt embedding
        :return: A numpy array of size #nodes * d
        """
        pass


class GCNModel(StaticGraphEmbedding):
    def __init__(self, params, method_name, all_graph, graph, file_tags, multilabel=False):
        super(GCNModel, self).__init__(params["dimension"], method_name, graph)
        # Training settings
        self._adj, self._features, self._labels, self._idx_train, self._idx_val, self._idx_test = \
            new_load_data(all_graph, graph, file_tags, len(graph.nodes()), params["nc_nodes"], multilabel=multilabel)
        self.multi = multilabel
        self._seed = 42
        self._epochs = params["epochs"]
        self._lr = params["lr"]
        self._weight_decay = params["weight_decay"]
        self._dropout = params["dropout"]
        if multilabel:
            self._loss = torch.nn.BCEWithLogitsLoss()
            nclass = int(self._labels.shape[1])
        else:
            self._loss = F.nll_loss
            nclass = graph.number_of_nodes()
        self._model = GCN(nfeat=self._features.shape[1], nhid=self._d, nclass=nclass, dropout=self._dropout,
                          multilabel=self.multi)
        self._optimizer = optim.Adam(self._model.parameters(), lr=self._lr, weight_decay=self._weight_decay)

    def learn_embedding(self):
        self._model.train()
        for epoch in range(self._epochs):
            output1 = train_(epoch, self._epochs, self._model, self._optimizer, self._loss, self._features,
                             self._labels, self._adj, self._idx_train, self._idx_val, self.multi)
        y = output1.detach().numpy()
        nodes = list(self._graph.nodes())
        self._dict_embedding = {nodes[i]: y[i, :] for i in range(len(nodes))}
        return self._dict_embedding


class GraphFactorization(StaticGraphEmbedding):
    """
    Graph Factorization factorizes the adjacency matrix with regularization.
    Args: hyper_dict (object): Hyper parameters.
    """

    def __init__(self, params, method_name, graph):
        super(GraphFactorization, self).__init__(params["dimension"], method_name, graph)
        """
        Initialize the GraphFactorization class
        Args: params:
            d: dimension of the embedding
            eta: learning rate of sgd
            regu: regularization coefficient of magnitude of weights
            max_iter: max iterations in sgd
            print_step: #iterations to log the prgoress (step%print_step)
        """
        self._eta = params["eta"]
        self._regu = params["regularization"]
        self._max_iter = params["max_iter"]
        self._print_step = params["print_step"]
        self._X = np.zeros(shape=(len(list(self._graph.nodes())), self._d))

    def get_f_value(self):
        """
        Get the value of f- the optimization function
        """
        nodes = list(self._graph.nodes())
        new_names = list(np.arange(0, len(nodes)))
        mapping = {}
        for i in new_names:
            mapping.update({nodes[i]: str(i)})
        H = nx.relabel.relabel_nodes(self._graph, mapping)
        f1 = 0
        for i, j, w in H.edges(data='weight', default=1):
            f1 += (w - np.dot(self._X[int(i), :], self._X[int(j), :])) ** 2
        f2 = self._regu * (np.linalg.norm(self._X) ** 2)
        return H, [f1, f2, f1 + f2]

    def learn_embedding(self):
        """
        Apply graph factorization embedding
        """
        t1 = time()
        node_num = len(list(self._graph.nodes()))
        self._X = 0.01 * np.random.randn(node_num, self._d)
        for iter_id in range(self._max_iter):
            my_f = self.get_f_value()
            count = 0
            if not iter_id % self._print_step:
                H, [f1, f2, f] = self.get_f_value()
                print('\t\tIter id: %d, Objective: %g, f1: %g, f2: %g' % (
                    iter_id,
                    f,
                    f1,
                    f2
                ))
            for i, j, w in H.edges(data='weight', default=1):
                if j <= i:
                    continue
                term1 = -(w - np.dot(self._X[int(i), :], self._X[int(j), :])) * self._X[int(j), :]
                term2 = self._regu * self._X[int(i), :]
                delPhi = term1 + term2
                self._X[int(i), :] -= self._eta * delPhi
            if count > 30:
                break
        t2 = time()
        projections = {}
        nodes = list(self._graph.nodes())
        new_nodes = list(H.nodes())
        for j in range(len(nodes)):
            projections.update({nodes[j]: self._X[int(new_nodes[j]), :]})
        # X is the embedding matrix and projections are the embedding dictionary
        return self._X, (t2 - t1), projections

    def get_embedding(self):
        return self._X


class HOPE(StaticGraphEmbedding):
    def __init__(self, params, method_name, graph):
        super(HOPE, self).__init__(params["dimension"], method_name, graph)
        """
        Initialize the HOPE class
        Args:
            d: dimension of the embedding
            beta: higher order coefficient
        """
        self._beta = params["beta"]

    def learn_embedding(self):
        """
        Apply HOPE embedding
        """
        A = nx.to_scipy_sparse_matrix(self._graph, format='csc')
        I = identity(self._graph.number_of_nodes(), format='csc')
        M_g = I - - self._beta * A
        M_l = self._beta * A
        # A = nx.to_numpy_matrix(self._graph)
        # M_g = np.eye(len(self._graph.nodes())) - self._beta * A
        # M_l = self._beta * A
        # S = inv(M_g).dot(M_l)
        S = np.dot(inv(M_g), M_l)

        u, s, vt = svds(S, k=self._d // 2)
        X1 = np.dot(u, np.diag(np.sqrt(s)))
        X2 = np.dot(vt.T, np.diag(np.sqrt(s)))
        self._X = np.concatenate((X1, X2), axis=1)

        p_d_p_t = np.dot(u, np.dot(np.diag(s), vt))
        eig_err = np.linalg.norm(p_d_p_t - S)
        print('SVD error (low rank): %f' % eig_err)

        # create dictionary of nodes
        nodes = list(self._graph.nodes())
        projections = {}
        for i in range(len(nodes)):
            y = self._X[i]
            y = np.reshape(y, newshape=(1, self._d))
            projections.update({nodes[i]: y[0]})
        # X is the embedding matrix, S is the similarity, projections is the embedding dictionary
        return projections, S, self._X, X1, X2

    def get_embedding(self):
        return self._X


class NODE2VEC(StaticGraphEmbedding):
    """
    Nod2Vec Embedding using random walks
    """

    def __init__(self, params, method_name, graph):
        super(NODE2VEC, self).__init__(params["dimension"], method_name, graph)
        """
        parameters:
        "walk_length" - Length of each random walk
        "num_walks" - Number of random walks from each source nodes
        "workers" - How many times repeat this process
        """
        self._walk_length = params["walk_length"]
        self._num_walks = params["num_walks"]
        self._workers = params["workers"]

    def learn_embedding(self):
        """
        Apply Node2Vec embedding
        """
        node2vec = Node2Vec(self._graph, dimensions=self._d, walk_length=self._walk_length,
                            num_walks=self._num_walks, workers=self._workers)
        model = node2vec.fit()
        nodes = list(self._graph.nodes())
        self._my_dict = {}
        for node in nodes:
            self._my_dict.update({node: np.asarray(model.wv.get_vector(node))})
        self._X = np.zeros((len(nodes), self._d))
        for i in range(len(nodes)):
            self._X[i, :] = np.asarray(model.wv.get_vector(nodes[i]))
        # X is the embedding matrix and projections are the embedding dictionary
        return self._X, self._my_dict

    def get_embedding(self):
        return self._X, self._my_dict


def final(G, method_name, params, file_tags=None):
    """
    Final function to apply state-of-the-art embedding methods
    :param G: Graph to embed
    :param method_name: state-of-the-art embedding algorithm
    :param params: Parameters dictionary according to the embedding method
    :return:
    """
    if method_name == "HOPE":
        t = time()
        embedding = HOPE(params, method_name, G)
        projections, S, _, X1, X2 = embedding.learn_embedding()
        X = embedding.get_embedding()
        elapsed_time = time() - t
    elif method_name == "node2vec":
        t = time()
        embedding = NODE2VEC(params, method_name, G)
        embedding.learn_embedding()
        X, projections = embedding.get_embedding()
        elapsed_time = time() - t
    elif method_name == "GF":
        t = time()
        embedding = GraphFactorization(params, method_name, G)
        _, _, projections = embedding.learn_embedding()
        X = embedding.get_embedding()
        elapsed_time = time() - t
    elif method_name == "GAE":
        t = time()
        embedding = GAE(params, method_name, G)
        X, projections = embedding.embedding_matrix, embedding.embedding_dict
        elapsed_time = time() - t
    elif method_name == "GCN":
        t = time()
        embedding = GCNModel(params, method_name, G, G, file_tags=file_tags)
        projections = embedding.learn_embedding()
        elapsed_time = time() - t
        X = None
    else:
        sys.exit("Chosen GEA is not valid. Please choose one of the following: node2vec, HOPE, GF, GEA. If you have"
                 "tags to your data, you can also choose GCN")
    return X, projections, elapsed_time



import math
import torch
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module


class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """
    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'



"""
FODGE - Fast Online Dynamic Graph Embedding

The package has 3 main functions:

1. FODGE:

    >>> from FODGE import FODGE
    >>> FODGE(name='catalano', graph_path=".", save_path=".", initial_method="node2vec", dim=128,
    >>>       epsilon=0.04, alpha_exist=0.2, beta=0.7, number=50)

2. link_prediction_1:

    >>> from FODGE import link_prediction_1
    >>> link_prediction_1(name='catalano', graph_path=".", save_path=".", initial_method="node2vec", dim=128,
    >>>                   epsilon=0.04, alpha=0.2, beta=0.7, number=50)

2. link_prediction_2:

    >>> from FODGE import link_prediction_2
    >>> link_prediction_2(name='catalano', graph_path=".", save_path=".", initial_method="node2vec", dim=128,
    >>>                   epsilon=0.04, alpha=0.2, beta=0.7, number=50)
"""

from .evaluation_tasks.temporal_link_prediction import *
from .evaluation_tasks.calculate_non_edges import *
from .evaluation_tasks.linear_regression_link_prediction import *
from .main_functions import link_prediction_1, link_prediction_2



"""
Loader functions for the datasets. You can add yours to here.

The function requires 2 inputs:
- name: Name of dataset as the name of the file without suffix
- path: Path to where the file is located

The functions return 2 outputs:
- dict_snapshots: Dictionary of snapshots, where keys are time stamps and values are list of edges occurring at this
                    time.
- dict_weights: Dictionary of weights according to dict_snapshots - keys are the time stamps and values are list of
                weight edges corresponding to the order of edges in dict_snapshots.

Note that if you change the name of the function "data_loader", you have to change it as well in the file "main.py" in
in line 51.
"""

from datetime import datetime
import os
import numpy as np
import calendar


def data_loader(name, path):
    """
    Data loader for all datasets used.
    :param name: Name of the file (str, without .txt)
    :param path: Path to where it is
    :return:
    """
    if name == "dblp":
        dict_snapshots, dict_weights = load_dblp(name, path)
    else:
        dict_snapshots, dict_weights, times = {}, {}, []
        with open(os.path.join(path, name + ".txt"), 'r') as filehandle:
            for line in filehandle:
                a = line.split(" ")
                if len(a) == 1:
                    a = line.split(",")
                if "%" in a[0]:
                    continue
                else:
                    node1 = a[0]
                    node2 = a[1]
                    if node1 == node2:
                        continue
                    try:
                        w = float(a[2])
                        t = a[3].split("\n")[0]
                    except:
                        w = 1.
                        t = a[2].split("\n")[0]
                    if t == '0':
                        date = int(t)
                    else:
                        x = datetime.fromtimestamp(0)
                        x = add_months(x, int(t) - 1)
                        times.append(x)
                        month = x.month
                        year = x.year
                        date = "{}.{}".format(month, year)
                    if dict_snapshots.get(date) is None:
                        dict_snapshots.update({date: [(node1, node2)]})
                    else:
                        dict_snapshots[date].append((node1, node2))
                    if dict_weights.get(date) is None:
                        dict_weights.update({date: [w]})
                    else:
                        dict_weights[date].append(w)
        times.sort()
        sorted_times = sort_keys(times)
        dict_snapshots = sort_dict_snapshots(dict_snapshots, sorted_times)
        if name == "facebook-wosn-wall":
            del dict_snapshots["10.2004"]
            del dict_snapshots["11.2004"]
            del dict_snapshots["12.2004"]
    return dict_snapshots, dict_weights


def load_dblp(name, path, is_weighted=False):
    """
    DBLP dataset
    """
    dict_snapshots = {}
    with open(os.path.join(path, name + ".txt"), 'r') as filehandle:
        for line in filehandle:
            a = line.split(" ")
            node1 = a[0]
            node2 = a[1]
            if is_weighted:
                w = float(a[2])
                t = a[3].split("\n")[0]
            else:
                w = 1.
                t = a[2].split("\n")[0]
            if dict_snapshots.get(t) is None:
                dict_snapshots.update({t: [(node1, node2, w)]})
            else:
                dict_snapshots[t].append((node1, node2, w))
    sorted_dict = dict(sorted(dict_snapshots.items(), key=lambda kv: kv[0]))
    keys = sorted_dict.keys()
    dict_weights = {key: [] for key in keys}
    new_dict = {key: [] for key in keys}
    for key in keys:
        for edge in sorted_dict[key]:
            dict_weights[key].append(edge[2])
            new_dict[key].append((edge[0], edge[1]))
    return new_dict, dict_weights


def sort_keys(times):
    """
    Function to sort the times
    """
    sorted_times = []
    for t in times:
        year = t.year
        month = t.month
        date = "{}.{}".format(month, year)
        if date in sorted_times:
            continue
        else:
            sorted_times.append(date)
    return sorted_times


def sort_dict_snapshots(dict_snapshots, times):
    """
    Sort dictionary of snapshots by time
    :param dict_snapshots:
    :param times:
    :return:
    """
    new_dict = {t: dict_snapshots[t] for t in times}
    return new_dict


def load_embedding(path, file_name):
    """
    Given a .npy file - embedding of a given graph. return the embedding dictionary
    :param path: Where this file is saved.
    :param file_name: The name of the file
    :return: Embedding dictionary
    """
    data = np.load(os.path.join(path, '{}.npy'.format(file_name)), allow_pickle=True)
    dict_embedding = data.item()
    return dict_embedding


def add_months(sourcedate, months):
    month = sourcedate.month - 1 + months
    year = sourcedate.year + month // 12
    month = month % 12 + 1
    day = min(sourcedate.day, calendar.monthrange(year, month)[1])
    return datetime(year, month, day)



import torch.nn as nn
import torch.nn.functional as F
from .layers import GraphConvolution


class GCN(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, multilabel):
        super(GCN, self).__init__()

        self.gc1 = GraphConvolution(nfeat, nhid)
        self.gc2 = GraphConvolution(nhid, nclass)
        self.dropout = dropout
        self.multi = multilabel

    def forward(self, x, adj):
        x = F.relu(self.gc1(x, adj))
        x = F.dropout(x, self.dropout, training=self.training)
        hid = x
        x = self.gc2(x, adj)
        if self.multi:
            return hid, x
        else:
            return hid, F.log_softmax(x, dim=1)



import torch
import torch.nn.functional as F
from torch.nn.modules.module import Module
from torch.nn.parameter import Parameter


class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, dropout=0., act=F.relu):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.dropout = dropout
        self.act = act
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.weight)

    def forward(self, input, adj):
        input = F.dropout(input, self.dropout, self.training)
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)
        output = self.act(output)
        return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'



"""
Main file to run the second temporal link prediction task (linear regression)
"""

from sklearn.metrics import f1_score, accuracy_score, roc_auc_score
from sklearn.linear_model import LogisticRegression
from ..fodge.load_data import *
import pandas as pd
from ..fodge.run_fodge import FODGE
from ..GEA.all_gea import *


class TemporalLinkPrediction2:
    """
    Class to run dynamic link prediction task.
    """

    def __init__(self, name, save_path, graph_path, func=data_loader, initial_method="node2vec", dim=128, epsilon=0.01,
                 alpha_exist=0., beta=0.99, number=0, test_ratio=0.2, val_ratio=0.2, file_tags=None):
        """
        Init function to initialize this class.
        :param name: Name of the dataset (string).
        :param graph_path: Path to where the dataset file is.
        :param save_path: Path to where to save the calculated embedding
        :param func: Loader function to load the dataset as a dictionary of snapshots. In order to create your own
                     load function, one should define it in the file load_data.py. (Name of the function)
        :param initial_method: Initial state-of-the-art algorithm to embed the first snapshot with. Options are
                               "node2vec", "HOPE", "GAE", "GF" and "GCN".
        :param dim: Embedding dimension (int)
        :param epsilon: The weight that is given to the second order neighbours.
        :param alpha_exist: Weight that is given to the previous changed existing nodes embeddings when they are
        recalculated (float between 0 and 1).
        :param beta: Rate of exponential decay of the edges weights through time
        :param number: How many vertices in the cumulative initial snapshot (choose a number where a 5-core exists)
        :param test_ratio: Test ratio for temporal link prediction task (float)
        :param val_ratio: Val ratio for temporal link prediction task (float)
        :param file_tags: If GCN GEA is used, then one should profile a file of tags
        """
        initial_t = time()
        # initialize the FODGE class
        self.DE = FODGE(name, graph_path, save_path, func=func, initial_method=initial_method, dim=dim, epsilon=epsilon,
                        alpha_exist=alpha_exist, beta=beta, number=number, file_tags=file_tags, mission="lp2")
        self.start_index = self.DE.index
        self.graph_lp = self.DE.create_weighted_graph_for_all_times()
        print("number of nodes: ", self.graph_lp.number_of_nodes())
        self.dict_snapshots = self.DE.dict_snapshots.copy()

        # calculate the embedding using FODGE
        self.full_dict_embeddings, self.dict_all_embeddings, self.total_time = self.DE.calculate_embedding()
        self.DE.full_dict_embeddings, self.DE.dict_all_embeddings = self.full_dict_embeddings, self.dict_all_embeddings
        self.DE.save_embedding(save_path, mission="lp2")

        t = time() - initial_t
        print(f"FODGE is done after {t} seconds. Starting temporal link prediction task")

        print("dict is ready")
        measure_list = ["Avg", "Had", "L1", "L2"]
        node2idx = dict(zip(list(self.graph_lp.nodes()), np.arange(self.graph_lp.number_of_nodes())))
        link_prediction_all_time(name, node2idx, self.DE.dict_snapshots, self.dict_all_embeddings,
                                 measure_list, test_ratio, val_ratio, 1-test_ratio-val_ratio)


def get_neg_edge_samples(pos_edges, edge_num, all_edge_dict, node_num, add_label=True):
    """
    Generate negative links (edges) in a graph
    :param pos_edges: ids of positive edges (numpy array)
    :param edge_num: Number of edges to choose
    :param all_edge_dict: Dict of all edges
    :param node_num: Number of nodes in the whole temporal network
    :param add_label: True for adding True/False to positive/negative edges, else False.
    """
    neg_edge_dict = dict()
    neg_edge_list = []
    cnt = 0
    while cnt < edge_num:
        from_id = np.random.choice(node_num)
        to_id = np.random.choice(node_num)
        if from_id == to_id:
            continue
        if (from_id, to_id) in all_edge_dict or (to_id, from_id) in all_edge_dict:
            continue
        if (from_id, to_id) in neg_edge_dict or (to_id, from_id) in neg_edge_dict:
            continue
        if add_label:
            neg_edge_list.append([from_id, to_id, 0])
        else:
            neg_edge_list.append([from_id, to_id])
        cnt += 1
    neg_edges = np.array(neg_edge_list)
    all_edges = np.vstack([pos_edges, neg_edges])
    return all_edges


def generate_edge_sample(node2idx, t, dict_snapshots, test_ratio, val_ratio, train_ratio):
    """
    Generate edge samples for train, val and test for a specific time
    :param node2idx: Dict where keys are names of nodes and values are their corresponding ids.
    :param t: Current time
    :param dict_snapshots: Dict where keys are times and each value is a list of edges occurring in this time.
    :param test_ratio: Test ratio for temporal link prediction task (float)
    :param val_ratio: Val ratio for temporal link prediction task (float)
    :param train_ratio: Train ratio for temporal link prediction task (float)
    """
    all_edge_dict = {}
    edge_list = []
    edges = dict_snapshots[t]
    g = nx.Graph()
    g.add_edges_from(edges)
    n = g.number_of_nodes()
    for e in edges:
        from_id, to_id = node2idx[e[0]], node2idx[e[1]]
        all_edge_dict.update({(to_id, from_id): 1})
        all_edge_dict.update({(from_id, to_id): 1})
        edge_list.append([from_id, to_id, 1])
        edge_list.append([to_id, from_id, 1])
    all_edges = np.array(edge_list)
    del edge_list
    edge_num = all_edges.shape[0]

    np.random.shuffle(all_edges)
    test_num = int(np.floor(edge_num * test_ratio))
    val_num = int(np.floor(edge_num * val_ratio))
    train_num = int(np.floor((edge_num - test_num - val_num) * train_ratio))

    val_edges = all_edges[: val_num]
    test_edges = all_edges[val_num: val_num + test_num]
    train_edges = all_edges[val_num + test_num: val_num + test_num + train_num]
    del all_edges

    train_edges = get_neg_edge_samples(train_edges, train_num, all_edge_dict, len(node2idx))
    test_edges = get_neg_edge_samples(test_edges, test_num, all_edge_dict, len(node2idx))
    val_edges = get_neg_edge_samples(val_edges, val_num, all_edge_dict, len(node2idx))

    df_train = pd.DataFrame(train_edges, columns=['from_id', 'to_id', 'label'])
    df_test = pd.DataFrame(test_edges, columns=['from_id', 'to_id', 'label'])
    df_val = pd.DataFrame(val_edges, columns=['from_id', 'to_id', 'label'])

    return df_train, df_val, df_test


def generate_edge_samples_all_times(node2idx, dict_snapshots, test_ratio, val_ratio, train_ratio):
    """
    Generate edge samples for train, val and test for a specific time for all times
    :param node2idx: Dict where keys are names of nodes and values are their corresponding ids.
    :param dict_snapshots: Dict where keys are times and each value is a list of edges occurring in this time.
    :param test_ratio: Test ratio for temporal link prediction task (float)
    :param val_ratio: Val ratio for temporal link prediction task (float)
    :param train_ratio: Train ratio for temporal link prediction task (float)
    """
    times = list(dict_snapshots.keys())
    list_dataframes = {}
    for t in times:
        i, j, k = generate_edge_sample(node2idx, t, dict_snapshots, test_ratio, val_ratio, train_ratio)
        list_dataframes.update({t: (i, j, k)})
    return list_dataframes


def get_edge_feature(node2idx, measure_list, edge_arr, embedding_arr, dict_all_embeddings, curr_t):
    """
    For all the edges in the given array, compute its value with the embeddings of the corresponding nodes according
    to the measures.
    :param node2idx: Dict where keys are names of nodes and values are their corresponding ids.
    :param measure_list: List of measures. One of 'L1', 'L2', 'Had', 'Avg'
    :param embedding_arr: Dict of embeddings of the nodes
    :param edge_arr: Relevant edges (numpy array)
    :param dict_all_embeddings: Dict of the embeddings at any given time. Keys are times and values are dicts of the
                                embeddings of all the nodes that appear specifically at this time.
    :param curr_t: Current time
    """
    idx2node = {i: n for n, i in node2idx.items()}
    feature_dict = dict()
    times = list(dict_all_embeddings.keys())
    time2index = {times[j]: j for j in range(len(times))}
    for measure in measure_list:
        assert measure in ['Avg', 'Had', 'L1', 'L2']
        feature_dict[measure] = []
    for i, edge in enumerate(edge_arr):
        from_id, to_id = idx2node[edge[0]], idx2node[edge[1]]
        if embedding_arr.get(from_id) is None or embedding_arr.get(to_id) is None:
            for t in times:
                if time2index[t] <= time2index[curr_t]:
                    continue
                else:
                    if len(set([from_id]).intersection(set(dict_all_embeddings[t]))) == 1 and \
                            len(set([to_id]).intersection(set(dict_all_embeddings[t]))) == 1:
                        new_embed = dict_all_embeddings[t].copy()
                        break
        else:
            new_embed = embedding_arr.copy()
        for measure in measure_list:
            if measure == 'Avg':
                feature_dict[measure].append((new_embed[from_id] + new_embed[to_id]) / 2)
            elif measure == 'Had':
                feature_dict[measure].append(new_embed[from_id] * new_embed[to_id])
            elif measure == 'L1':
                feature_dict[measure].append(np.abs(new_embed[from_id] - new_embed[to_id]))
            elif measure == 'L2':
                feature_dict[measure].append((new_embed[from_id] - new_embed[to_id]) ** 2)
    for measure in measure_list:
        feature_dict[measure] = np.array(feature_dict[measure])
    return feature_dict


def train(node2idx, measure_list, train_edges, val_edges, embeddings, dict_all_embeddings, curr_t):
    """
    Train function!
    """
    print('Start training!')
    train_edges.columns = range(train_edges.shape[1])
    val_edges.columns = range(val_edges.shape[1])
    train_labels = train_edges.iloc[:, -1].values
    val_labels = val_edges.iloc[:, -1].values
    train_edges = train_edges.values
    val_edges = val_edges.values
    train_feature_dict = get_edge_feature(node2idx, measure_list, train_edges, embeddings, dict_all_embeddings, curr_t)
    val_feature_dict = get_edge_feature(node2idx, measure_list, val_edges, embeddings, dict_all_embeddings, curr_t)

    model_dict = dict()
    for measure in measure_list:
        if measure == 'sigmoid':
            continue
        models = []
        # for C in [0.01, 0.1, 1, 10]:
        for C in [0.01, 0.1, 1, 10]:
            model = LogisticRegression(C=C, solver='lbfgs', max_iter=1000, class_weight='balanced')
            model.fit(train_feature_dict[measure], train_labels)
            models.append(model)
        best_auc = 0
        model_idx = -1
        for i, model in enumerate(models):
            val_pred = model.predict_proba(val_feature_dict[measure])[:, 1]
            auc = roc_auc_score(val_labels, val_pred)
            if auc >= best_auc:
                best_auc = auc
                model_idx = i
        model_dict[measure] = models[model_idx]
    print('Finish training!')
    return model_dict


def test_function(node2idx, measure_list, test_edges, embeddings, model_dict, date, dict_all_embeddings):
    """
    Test function!
    """
    test_labels = test_edges.iloc[:, -1].values
    test_edges = test_edges.values
    test_feature_dict = get_edge_feature(node2idx, measure_list, test_edges, embeddings, dict_all_embeddings, date)
    auc_list = [date]
    for measure in measure_list:
        if measure == 'sigmoid':
            test_pred = test_feature_dict[measure]
        else:
            test_pred = model_dict[measure].predict_proba(test_feature_dict[measure])[:, 1]
        auc_list.append(roc_auc_score(test_labels, test_pred))
    return auc_list


def link_prediction_all_time(name, node2idx, dict_snapshots, dict_all_embeddings, measure_list, test_ratio, val_ratio,
                             train_ratio):
    """
    Final function to perform temporal link prediction task
    """
    dataframes = generate_edge_samples_all_times(node2idx, dict_snapshots, test_ratio, val_ratio, train_ratio)
    times = list(dataframes.keys())
    all_auc_list = []

    for t in times:
        train_edges, val_edges, test_edges = dataframes[t]
        print('Current date is: {}'.format(t))
        embeddings = dict_all_embeddings[t]
        model_dict = train(node2idx, measure_list, train_edges, val_edges, embeddings, dict_all_embeddings, t)
        auc_list = test_function(node2idx, measure_list, test_edges, embeddings, model_dict, t, dict_all_embeddings)
        all_auc_list.append(auc_list)
    # print('all auc list len: ', len(all_auc_list)
    column_names = ['date'] + measure_list
    df_output = pd.DataFrame(all_auc_list, columns=column_names)
    print(df_output)
    print('method = FODGE', ', average AUC of Had: ', df_output["Had"].mean())
    print('method = FODGE', ', average AUC of Avg: ', df_output["Avg"].mean())
    print('method = FODGE', ', average AUC of L1: ', df_output["L1"].mean())
    print('method = FODGE', ', average AUC of L2: ', df_output["L2"].mean())

    # save the results in a csv file
    output_file_path = os.path.join("lp_results_{}.csv".format(name))
    df_output.to_csv(output_file_path, sep=',', index=False)



"""
Main functions to run FODGE.
"""
from .evaluation_tasks.temporal_link_prediction import *
from .evaluation_tasks.calculate_non_edges import *
from .evaluation_tasks.linear_regression_link_prediction import *


def link_prediction_1(name, save_path, graph_path, func=data_loader, non_edges_file=None,
                      initial_method='node2vec', dim=128, epsilon=0.01,
                      alpha=0.2, beta=0.3, number=1000,
                      test_ratio=0.2, file_tags=None):
    if non_edges_file is None:  # create non_edges_file if the user does not give one
        path_non_edges_file = create_non_edges_file(name, graph_path, func)
    else:
        path_non_edges_file = os.path.join(graph_path, non_edges_file)
    TemporalLinkPrediction1(name, graph_path, save_path, func=func,
                            initial_method=initial_method, dim=dim, epsilon=epsilon,
                            alpha_exist=alpha, beta=beta, number=number, test_ratio=test_ratio,
                            non_edges_file=path_non_edges_file, file_tags=file_tags)


def link_prediction_2(name, save_path, graph_path, func=data_loader,
                      initial_method='node2vec', dim=128, epsilon=0.01,
                      alpha=0.2, beta=0.3, number=1000,
                      test_ratio=0.2, val_ratio=0.3, file_tags=None):
    TemporalLinkPrediction2(name, save_path, graph_path, func=func,
                            initial_method=initial_method, dim=dim, epsilon=epsilon,
                            alpha_exist=alpha, beta=beta, number=number, test_ratio=test_ratio,
                            val_ratio=val_ratio, file_tags=file_tags)



import torch
import torch.nn as nn
import torch.nn.functional as F

from .layers import GraphConvolution


class GCNModelVAE(nn.Module):
    def __init__(self, input_feat_dim, hidden_dim1, hidden_dim2, dropout):
        super(GCNModelVAE, self).__init__()
        self.gc1 = GraphConvolution(input_feat_dim, hidden_dim1, dropout, act=F.relu)
        self.gc2 = GraphConvolution(hidden_dim1, hidden_dim2, dropout, act=lambda x: x)
        self.gc3 = GraphConvolution(hidden_dim1, hidden_dim2, dropout, act=lambda x: x)
        self.dc = InnerProductDecoder(dropout, act=lambda x: x)

    def encode(self, x, adj):
        hidden1 = self.gc1(x, adj)
        return self.gc2(hidden1, adj), self.gc3(hidden1, adj)

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(logvar)
            eps = torch.randn_like(std)
            return eps.mul(std).add_(mu)
        else:
            return mu

    def forward(self, x, adj):
        mu, logvar = self.encode(x, adj)
        z = self.reparameterize(mu, logvar)
        return self.dc(z), mu, logvar, z


class InnerProductDecoder(nn.Module):
    """Decoder for using inner product for prediction."""

    def __init__(self, dropout, act=torch.sigmoid):
        super(InnerProductDecoder, self).__init__()
        self.dropout = dropout
        self.act = act

    def forward(self, z):
        z = F.dropout(z, self.dropout, training=self.training)
        adj = self.act(torch.mm(z, z.t()))
        return adj



"""
Utils file of FODGE
"""

from ..GEA.all_gea import *
import time as my_time
from scipy.linalg import orthogonal_procrustes
import numpy as np
import networkx as nx


def user_print(item, user_wish):
    """
    a function to show the user the state of the code. If you want a live update of the current state of the code and
    some details: set user wish to True else False
    """
    if user_wish is True:
        print(item, sep=' ', end='', flush=True)
        my_time.sleep(3)
        print(" ", end='\r')


def get_initial_proj_nodes_by_k_core(G, number):
    """
    Function to decide which nodes would be in the initial embedding by k-core score.
    :param G: Our graph
    :param number: Controls number of nodes in the initial projection
    :return: A list of the nodes that are in the initial projection
    """
    G.remove_edges_from(nx.selfloop_edges(G))
    core_dict = nx.core_number(G)
    sorted_core_dict = {k: v for k, v in sorted(core_dict.items(), key=lambda item: item[1], reverse=True)}
    keys = list(sorted_core_dict.keys())
    chosen_nodes = keys[:number]
    return chosen_nodes


def jaccard_similarity(list1, list2):
    """
    Calculate Jaccard Similarity between two sets
    """
    intersection = len(list(set(list1).intersection((set(list2)))))
    union = len(list1) + len(list2) - intersection
    return float(intersection) / union


def from_dict_to_matrix(my_dict, d):
    """
    Convert a dictionary of embeddings (key is the vertex id and value is its embedding of dimension d) to a matrix
    of size nxd where n is number of nodes and d is the embedding dimension.
    """
    values = list(my_dict.values())
    for i in range(len(values)):
        v = values[i]
        t = v.copy()
        t = np.reshape(t, (d, 1))
        values[i] = t
    matrix = np.concatenate(tuple(values), axis=1)
    return matrix


def dict_of_core(core_nodes, dict_proj):
    """
    Create the dict of embeddings of the core nodes only
    """
    new_dict = {core_nodes[i]: dict_proj[core_nodes[i]] for i in range(len(core_nodes))}
    return new_dict


def find_rotation_matrix(p, q):
    """
    Find a rotation matrix using orthogonal procrustes
    """
    assert p.shape == q.shape
    n, dim = p.shape

    center_p = p - p.mean(axis=0)
    center_q = q - q.mean(axis=0)

    C = np.dot(np.transpose(center_p), center_q) / n
    V, S, W = np.linalg.svd(C)
    d = (np.linalg.det(V) * np.linalg.det(W)) < 0.0

    if d:
        S[-1] = -S[-1]
        V[:, -1] = -V[:, -1]

    R = np.dot(V, W)

    varP = np.var(p, axis=0).sum()
    c = 1 / varP * np.sum(S)

    t = q.mean(axis=0) - p.mean(axis=0).dot(c * R)

    return c, R, t


def rotation(dict_core_i_1, dict_core_i):
    """
    Perform the rotation of the new core based on it and the previous one
    """
    keys_i = set(dict_core_i.keys())
    keys_i_1 = set(dict_core_i_1.keys())
    nodes = list(keys_i.intersection(keys_i_1))
    print("common core is: ", len(nodes))
    Q_t_1 = np.array([dict_core_i[node] for node in nodes])
    Q_t = np.array([dict_core_i_1[node] for node in nodes])
    R_t, _ = orthogonal_procrustes(Q_t, Q_t_1)
    Q_t = np.array([dict_core_i_1[node] for node in dict_core_i_1])
    R_tQ_t = np.dot(Q_t, R_t)
    new_dict = {node: vec for node, vec in zip(dict_core_i_1, R_tQ_t)}
    return new_dict


def add_weights(G, weights):
    """
    Given a graph and a list of weights, add the weights to each edge of the graph (the order of the list of weights
    must be corresponding to the list of edges in the given graph)/
    :param G: The given graph
    :param weights: A list of weights (floats).
    :return: The graph where attribute of "weight" for each edge is added.
    """
    edges = list(G.edges())
    for i in range(len(edges)):
        e = edges[i]
        # G[e[0]][e[1]] = {"weight": weights[i]}
        G[e[0]][e[1]]["weight"] = 1
    return G


def divide_snapshot_nodes(previous_nodes, next_nodes):
    """
    For each snapshot divide the nodes into three groups: new- nodes that weren't in the previous snapshot,
    exist- nodes that are in both sequential snapshots, disappear- nodes that were in the previous snapshot but in the
    next they disappeared.
    :param previous_nodes: Node from snapshot t-1 (list)
    :param next_nodes: Nodes from snapshot t (list)
    :return: 3 lists of nodes as explained above
    """
    set_previous = set(previous_nodes)
    set_next = set(next_nodes)
    new = list(set_next - set_previous)
    exist = list(set_next.intersection(set_previous))
    disappear = list(set_previous - set(exist))
    return new, exist, disappear


def calculate_first_snapshot_embedding(user_wish, g_list_, nodes_list_, initial_method, params, file_tags=None):
    """
    Function to calculate the embedding of the first snapshot with state-of-the-art embedding method.
    :param user_wish: True if the user wants useful things to be printed, else False.
    :param g_list_: The list of graphs- graph for each time stamp.
    :param nodes_list_: List of lists of nodes- each time stamp has its own list of nodes
    :param initial_method: State-of-the-srt embedding method to embed the graph with.
    :param params: Parameters dict corresponding to the initial embedding used
    :return: The graph representing the first snapshot, initial dict of embeddings and initial set of nodes that are
             currently in the embedding.
    """
    # start with the first snapshot
    first_snapshot = g_list_[0]
    n = len(nodes_list_[0])
    e = first_snapshot.number_of_edges()
    user_print("number of nodes in first snapshot is: " + str(n), user_wish)
    user_print("number of edges in first snapshot is: " + str(e), user_wish)

    first_snapshot_nodes = nodes_list_[0]
    set_G_nodes = set(first_snapshot_nodes)

    # calculate the first snapshot embedding with a state-of-the-art algorithm
    user_print("calculate the projection of the first graph with {}...".format(initial_method), user_wish)
    _, dict_projections, _ = final(first_snapshot, initial_method, params, file_tags=file_tags)

    return first_snapshot, dict_projections, set_G_nodes


def initial_function(initial_method, g_list, nodes_list, params, file_tags=None):
    """
    Initial function to calculate the initial embedding, i.e. the embedding of the first snapshot.
    :param initial_method: embedding algorithm to embed the graph with
    :param g_list: The list of graph snapshots
    :param nodes_list: List of nodes for each time stamp
    :param file_tags: File of tags for each vertex if GEA used is GCN, else False
    :param params: Parameters for the embedding method
    :return: Initial embedding dictionary
    """
    user_wish = True

    # calculate fist snapshot embedding
    first_snapshot, dict_projections, set_proj_nodes = calculate_first_snapshot_embedding(user_wish, g_list,
                                                                                          nodes_list,
                                                                                          initial_method, params,
                                                                                          file_tags=file_tags)
    return dict_projections


def create_two_embedding_dicts(dict_snapshots, dict_projections):
    """
    First initialization of the two wanted embedding dictionaries. They are first embed with the nodes of the first
    snapshot and their embeddings.
    :param dict_snapshots: Dict where keys are times and values are a list of edges for each time stamp.
    :param dict_projections: Dict embeddings- currently keys are the nodes of the first snapshot and the values are
                             their embeddings in the first time stamp.
    :return: - full_dict_embeddings - A dictionary of all nodes embedding from all snapshots. If a node shows up in more
               than one snapshot so its final embedding in this dictionary is its embedding in the last time stamp the
               node has been shown,
             - dict_all_embeddings: A dictionary of dictionary. The keys are times and the value for each key is the
               embedding dictionary for the specific time.
             - times- the list of time stamps.
    """
    times = list(dict_snapshots.keys())

    dict_all_embeddings = {}
    dict_all_embeddings.update({times[0]: dict_projections.copy()})

    full_dict_embeddings = dict_projections.copy()

    return times, dict_all_embeddings, full_dict_embeddings


def create_dict_neighbors(H):
    """
    Given our undirected graph, Create a dictionary where value==node and key==set of its neighbours.
    """
    G = H.to_undirected()
    G_nodes = list(G.nodes())
    neighbors_dict = {}
    for i in range(len(G_nodes)):
        node = G_nodes[i]
        neighbors_dict.update({node: set(G[node])})
    return neighbors_dict


def create_dicts_of_connections(set_proj_nodes, set_no_proj_nodes, neighbors_dict):
    """
    a function that creates 3 dictionaries:
    1. dict_node_node (explained below)
    2. dict_node_enode (explained below)
    2. dict_enode_enode (explained below)
    """
    # value == (node that isn't in the embedding), key == (set of its neighbours that are also not in the embedding)
    dict_node_node = {}
    # value == (node that isn't in the embedding), key == (set of neighbours thar are in the embedding)
    dict_node_enode = {}
    # key==(node that is in the projection and has neighbors in it), value==(set of neighbors that are in projection)
    dict_enode_enode = {}
    # nodes that are not in the projection
    list_no_proj = list(set_no_proj_nodes)
    list_proj = list(set_proj_nodes)
    # not_connected_to_embed = []
    for i in range(len(list_no_proj)):
        node = list_no_proj[i]
        # neighbors of the node that aren't in the projection
        set1 = neighbors_dict[node].intersection(set_no_proj_nodes)
        dict_node_node.update({node: set1})
        # neighbors of the node that are in the projection
        set2 = neighbors_dict[node].intersection(set_proj_nodes)
        if len(set2) > 0:
            dict_node_enode.update({node: set2})
    for i in range(len(list_proj)):
        node = list_proj[i]
        # neighbors of the node that are in the projection
        set1 = neighbors_dict[node].intersection(set_proj_nodes)
        if len(set1) > 0:
            dict_enode_enode.update({node: set1})
    return dict_node_node, dict_node_enode, dict_enode_enode


def create_mapping(dict_times):
    """
    If times are not integers, transform them into integers.
    :param dict_times: Dict where keys are times.
    :return: A mapping which is a dictionary maps from current names of time stamps to integers (by their index)
    """
    keys = list(dict_times.keys())
    mapping = {}
    for i in range(len(keys)):
        mapping.update({keys[i]: i})
    return mapping


def create_new_cc(nodes_list, G, to_undirected=False):
    """
    If there are new nodes in the current snapshots that do not connect with nodes that were in the previous snapshot
    it means they create a new connected component. This function returns the sub graph that is built by these nodes
    and their connections with each other.
    :param nodes_list: A list of nodes in the new sub graph
    :param G: The given graph in the current snapshot
    :param to_undirected: If the graph is directed it needs to first be undirected.
    :return: The sub graph
    """
    if to_undirected is True:
        H = G.to_undirected()
        sub_G = nx.subgraph(H, nodes_list)
    else:
        sub_G = nx.subgraph(G, nodes_list)
    return sub_G


def create_new_connected_component(dict_projections, dict_cc, dict_nodes_cc, g_list_, set_no_proj, initial_method,
                                   params, i, file_tags=None):
    """
    If needed, create new connect component and update wanted dicts.
    :param dict_projections: Embedding dict
    :param dict_cc: Dict where keys are the number of the connected component and values are list of nodes that are in
                    this cc.
    :param dict_nodes_cc: Dict where keys are nodes and values is the number representing the cc they are in.
    :param g_list_: List of graphs for each time stamp.
    :param set_no_proj: Set of nodes that are currently not in the embedding because they create together a new cc.
    :param initial_method: State-of-the-art method to embed them with.
    :param params: Dict of parameters corresponding to the initial method.
    :param i: Index of the time stamp
    :param file_tags: If GCN GEA is used, one needs to provide file of tags
    :return: Updated dict_cc, dict_nodes_cc, and embedding dictionary.
    """
    new_cc = create_new_cc(list(set_no_proj), g_list_[i + 1], to_undirected=True)
    dict_cc, dict_nodes_cc = add_new_cc(new_cc, dict_nodes_cc, dict_cc)
    if new_cc.number_of_nodes() < params["dimension"] and initial_method == "HOPE":
        dim = params["dimension"]
        initial_method = "node2vec"
        params = {"dimension": dim, "walk_length": 80, "num_walks": 16, "workers": 2}
    _, dict_proj_new_cc, _ = final(new_cc, initial_method, params, file_tags=file_tags)
    z = {**dict_projections, **dict_proj_new_cc}.copy()
    return dict_cc, dict_nodes_cc, z


def create_dicts_cc(first_snapshot):
    """
    Function to create two important dictionaries:
    - dict_cc : keys are the different numbers representing the different connected components in the graph , while
    each value is a list of the nodes that are in the corresponding connected component.
    - dict_nodes_cc: keys are the nodes and each value is the connected component the node is related to.
    :param first_snapshot: The first snapshot graph
    :return: the two dictionaries
    """
    list_cc = sorted(nx.connected_components(first_snapshot), key=len, reverse=True)
    number_cc = len(list_cc)
    dict_cc = {i: list(c) for i, c in zip(list(range(number_cc)), list_cc)}
    dict_nodes_cc = {}
    for j in range(len(list_cc)):
        dict_nodes_cc.update({i: j for i in list_cc[j]})
    return dict_cc, dict_nodes_cc


def list_first_second_neigh(node, dict_neighbours):
    """
    Find the list of first and second nodes
    :param node: A node in the graph
    :param dict_neighbours: dict where keys is a node and value is a set of its node
    :return: Set of first and second neighbours
    """
    neigh = dict_neighbours[node].copy()
    for n in list(neigh):
        set1 = dict_neighbours[n].copy()
        neigh = neigh.union(set1)
    return neigh


def check_changed_existing(nodes_exist, dict_neighbors_old, dict_neighbors_new, dict_projections, second=False):
    """
    A function to check which nodes out of the nodes that are common to the current and previous snapshot have changed
    their neighbours. If they have, their embedding should be recalculated as new ones, but their current embeddings
    should be preserved for later use. In addition, they are deleted from the embedding dictionary (for now).
    :param nodes_exist: Common nodes between current and precious snapshots.
    :param dict_neighbors_old: Dictionary of neighbours of previous snapshots
    :param dict_neighbors_new: Dictionary of neighbours of current snapshots
    :param dict_projections: Dictionary of embeddings
    :param second: True if considering both first and second order neighbors, else False
    :return: A list of all changed nodes, a dictionary of the embeddings of all changed nodes, the update embedding
    dictionary of all nodes.
    """
    changed_exist = []
    changed_exist_proj_dict = {}
    for node in nodes_exist:
        if dict_projections.get(node) is not None:
            # changed existing nodes are nodes that their first neighbours are changed (don't check second order neighbors)
            if second is False:
                if dict_neighbors_old[node] == dict_neighbors_new[node]:
                    continue
                else:
                    changed_exist.append(node)
                    changed_exist_proj_dict.update({node: dict_projections[node]})
                    del dict_projections[node]
            # changed existing nodes are nodes that their first or second neighbours are changed
            else:
                first_second_neighbors_old = list_first_second_neigh(node, dict_neighbors_old)
                first_second_neighbors_new = list_first_second_neigh(node, dict_neighbors_new)
                if first_second_neighbors_old == first_second_neighbors_new:
                    continue
                else:
                    changed_exist.append(node)
                    changed_exist_proj_dict.update({node: dict_projections[node]})
                    del dict_projections[node]
    return changed_exist, changed_exist_proj_dict, dict_projections


def add_new_cc(graph, dict_nodes_cc, dict_cc):
    """
    If a new connected component is added in the current snapshot, update dict_cc and dict_nodes_cc.
    :param graph: The new sub graph that is added, it has its own connected components.
    :param dict_nodes_cc: Explained earlier
    :param dict_cc: Explained earlier
    :return: Two updated dicts
    """
    list_cc = sorted(nx.connected_components(graph), key=len, reverse=True)
    number_cc_new = len(list_cc)
    last_cc = sorted(list(dict_cc.keys()), reverse=False)[-1] + 1
    dict_cc.update({last_cc + j: list(list_cc[j]) for j in range(number_cc_new)})
    for k in range(len(list_cc)):
        dict_nodes_cc.update({i: last_cc + k for i in list_cc[k] if dict_nodes_cc.get(i) is None})
    return dict_cc, dict_nodes_cc


def update_full_dict_projections(nodes, full_dict_projections, t_dict_proj):
    """
    Full dict embeddings is a dictionary where keys are nodes and values are their embedding in the latest time stamp
    they have shown up, i.e. if a node is in oth time stamps t and k where t < k, then its embedding here is of
    time k (the bigger one).
    :param nodes: Nodes of current snapshot
    :param full_dict_projections: The current full_dict_projections
    :param t_dict_proj: embedding dict of time t
    :return: Updated full_dict_projections
    """
    for node in nodes:
        a = t_dict_proj[node]
        if full_dict_projections.get(node) is None:
            full_dict_projections.update({node: a})
        else:
            full_dict_projections[node] = a
    return full_dict_projections


def update_dicts_cc(node, dict_node_enode, dict_cc, dict_nodes_cc):
    """
    For a new node added to the embedding, dicts of connected components should be updated.
    :param node: Current node
    :param dict_node_enode: Dict where value == (node that isn't in the embedding), key == (set of neighbours thar are
                            in the embedding)
    :param dict_cc: Dict where keys are the number of the connected component and values are list of nodes that are in
                    this cc.
    :param dict_nodes_cc: Dict where keys are nodes and values is the number representing the cc they are in.
    :return: Updated dict_cc, dict_nodes_cc and list_cc which is the list of cc that needs to be united together.
    """
    list_cc = []
    neighbors = dict_node_enode[node]
    values = [dict_nodes_cc[n] for n in neighbors]
    values = list(dict.fromkeys(values))
    values.sort()
    # the node does not connect between two connected components

    if dict_nodes_cc.get(node) is None:
        dict_nodes_cc.update({node: values[0]})
        dict_cc[values[0]].append(node)
    else:
        current_cc = dict_nodes_cc[node]
        if current_cc != values[0]:
            dict_nodes_cc.update({node: values[0]})
            dict_cc[values[0]].append(node)
            my_set = set(dict_cc[current_cc])
            my_set2 = my_set - set([node])
            if len(my_set2) == 0:
                del dict_cc[current_cc]
            else:
                dict_cc[current_cc] = list(my_set2)

    if len(values) > 1:
        for i in values:
            list_cc.append(i)
    # the node connects between two connected components, it will be in the earliest cc

    return list_cc, dict_nodes_cc, dict_cc


def find_removal_edges(g1, g2):
    """
    Given a graph in snapshot t and another graph in snapshot t+1, find the removal edges, i.e. the ones that in G_t but
    not in G_(t+1).
    :param g1: Graph in snapshot t
    :param g2: Graph in snapshot t+1
    :return: List of removal edges
    """
    edges_1 = set(g1.edges())
    new_edges_1 = []
    for e in edges_1:
        new_edges_1.append(e if int(e[0]) < int(e[1]) else (e[1], e[0]))
    edges_2 = set(g2.edges())
    new_edges_2 = []
    for e in edges_2:
        new_edges_2.append(e if int(e[0]) < int(e[1]) else (e[1], e[0]))
    return list(set(new_edges_1) - set(new_edges_2))


def find_changed_nodes(removal_edges, g):
    """
    Find the nodes that are changed, i.e the nodes that are common to both consecutive snapshots and their neighbors are
    changed.
    :param removal_edges: List of removal edges
    :param g: The previous snapshot
    :return: List of changed existing nodes.
    """
    nodes = set(g.nodes())
    nodes_list = []
    for e in removal_edges:
        nodes_list.append(e[0])
        nodes_list.append(e[1])
    nodes_list = set(nodes_list).intersection(nodes)
    return nodes_list


def calculate_union_embedding(g, dict_projections, changed_nodes, nodes, nodes_previous, initial_method, params):
    """
    Calculate the embeddings of the united connected components
    """
    _, dict_enode_proj, _ = final(g, initial_method, params)
    non_changed_nodes = set(nodes) - changed_nodes
    dict_projections.update({n: dict_enode_proj[n] for n in non_changed_nodes})
    changed_previous = changed_nodes.intersection(set(nodes_previous))
    if len(changed_previous) > 0:
        for l in changed_nodes:
            del dict_projections[l]
    dict_changed_proj = {c: dict_enode_proj[c] for c in changed_nodes}
    # del dict_projections[l] for l in changed_nodes if l in nodes_previous
    return dict_projections, non_changed_nodes, dict_changed_proj


def create_dict_cc_nodes_cc(dict_projections):
    """
    Create dictionary of connected components.
    :param dict_projections: Dict of nodes and their embeddings
    :return: dict cc: key - the number of the cc , value - list of nodes in this cc
             dict_nodes_cc: key - the graph's node , value - the number of cc the node is in
    """
    keys = list(dict_projections.keys())
    dict_cc = {0: keys}
    dict_nodes_cc = {m: 0 for m in keys}
    return dict_cc, dict_nodes_cc


def changes_existing_nodes(nodes_exist, alpha, dict_proj, dict_proj_changed_exist):
    """
    For nodes that are in both sequential snapshots, update their embedding by their current embedding and new
    calculated embedding. Do it only for those that their neighbours are changed.
    :param nodes_exist: Nodes that exist in both sequential snapshots and their neighbours are changed
    :param alpha: Parameter representing the importance that is given to the old embedding
    :param dict_proj: Dictionary of embeddings of the nodes
    :param dict_proj_changed_exist: Dict of previous embedding of the changed existing nodes.
    :return: An updated embedding dictionary
    """
    for node in nodes_exist:
        if dict_proj.get(node) is not None:
            new_proj = dict_proj[node]
            old_proj = dict_proj_changed_exist[node]
            final_proj = alpha * old_proj + (1 - alpha) * new_proj
            dict_proj[node] = final_proj
    return dict_proj



"""
File to generate non-edges csv file
"""

import itertools as IT
import csv
import json
import random
import os
import networkx as nx
from datetime import datetime
from ..fodge.load_data import *


def create_graphs(dict_snapshots):
    """
    From dictionary of times and edges for each time, returen a list of static graphs for each snapshot, list of
    nodes of each graph and list of remaining nodes.
    """
    keys = list(dict_snapshots.keys())
    g_list = []
    for i in range(len(keys)):
        G = nx.DiGraph()
        G.add_edges_from(dict_snapshots[keys[i]])
        H = G.to_undirected()
        g_list.append(H.copy())
        print(i)
    return g_list


def load_graph(func, t):
    """
    Function to load the graph as a dictionary of snapshots
    :param func:
    :param t:
    :return:
    """
    dict_snapshots, dict_weights = func(*t)
    return dict_snapshots, dict_weights


def create_non_edges_file(name, path, func):
    print("First creating non edges file")
    dict_snapshots, dict_weights = load_graph(func, (name, path))
    print("Starting create graphs")
    g_list = create_graphs(dict_snapshots)
    my_list = []
    print("Starting find non edges for each time")
    for i in range(len(g_list)):
        g = g_list[i]
        e = g.number_of_edges()
        n = g.number_of_nodes()
        print(i, n, e)
        nodes = list(g.nodes())
        # could be very big so change 0.5 to smaller values if needed.
        indexes = random.sample(range(0, len(nodes)), int(len(nodes) * 0.5))
        new_nodes = []
        for j in indexes:
            new_nodes.append(nodes[j])
        missing = [pair for pair in IT.combinations(new_nodes, 2) if not g.has_edge(*pair)]
        print(i, len(missing))
        for pair in missing:
            my_list.append((i, pair[0], pair[1]))

    # indexes = random.sample(range(0, len(my_list)), int(len(my_list)))
    # new_list = []
    # for l in indexes:
    #   new_list.append(my_list[l])
    print("Almost done, writing the csv file!")
    non_edges_path = os.path.join(path, f"non_edges_{name}.csv")
    csvfile = open(non_edges_path, 'w', newline='')
    obj = csv.writer(csvfile)
    obj.writerows(my_list)
    csvfile.close()
    print("Non edges file is ready!")

    return non_edges_path


# name_ = "facebook_friendships"
# datasets_path = os.path.join("..", "datasets")
# func_ = data_loader
# create_non_edges_file(name_, datasets_path, func_)



"""
Main file to run the second temporal link prediction task (neural network)
"""


from sklearn.metrics import f1_score, accuracy_score, roc_auc_score
from ..fodge.load_data import *
from ..fodge.run_fodge import FODGE
from keras.layers import Input, Dense, Activation
from keras.models import Model, Sequential
from ..GEA.all_gea import *
import keras
import tensorflow as tf
from sklearn.metrics import roc_curve
from sklearn.metrics import auc
from ..evaluation_tasks.scores_utils import *
from ..evaluation_tasks.tlp_utils import *


class TemporalLinkPrediction1:
    """
    Class to run dynamic link prediction task.
    """
    def __init__(self, name, graph_path, save_path, func=data_loader, initial_method="node2vec", dim=128,
                 epsilon=0.01, alpha_exist=0., beta=0.7, number=0, test_ratio=0.2, non_edges_file="non_edges_dblp.csv",
                 file_tags=None):
        """
        Init function to initialize this class.
        :param name: Name of the dataset (string).
        :param graph_path: Path to where the dataset file is.
        :param save_path: Path to where to save the calculated embedding
        :param func: Loader function to load the dataset as a dictionary of snapshots. In order to create your own
                     load function, one should define it in the file load_data.py. (Name of the function)
        :param initial_method: Initial state-of-the-art algorithm to embed the first snapshot with. Options are
                               "node2vec", "HOPE", "GAE", "GF" and "GCN".
        :param dim: Embedding dimension (int)
        :param epsilon: The weight that is given to the second order neighbours.
        :param alpha_exist: Weight that is given to the previous changed existing nodes embeddings when they are
        recalculated (float between 0 and 1).
        :param beta: Rate of exponential decay of the edges weights through time
        :param number: How many vertices in the cumulative initial snapshot (choose a number where a 5-core exists)
        :param test_ratio: Test ratio for temporal link prediction task (float)
        :param non_edges_file: CSV file containing non edges. Can be created with the file 'calculate_non_edges.py'
        :param file_tags: If GCN GEA is used, then one should profile a file of tags
        """
        initial_t = time()
        # initialize the FODGE class
        self.DE = FODGE(name, graph_path, save_path, func=func, initial_method=initial_method, dim=dim, epsilon=epsilon,
                        alpha_exist=alpha_exist, beta=beta, number=number, file_tags=file_tags, mission="lp1")
        self.start_index = self.DE.index
        self.graph = self.DE.create_weighted_graph_for_all_times()
        self.dict_snapshots = self.DE.dict_snapshots.copy()

        # prepare data for temporal link prediction task, including finding the train and test edges changing the
        # dict of snapshots accordingly
        self.index, self.nodes, self.DE.dict_snapshots, self.true_edges_train, self.true_edges_test, self.K_train, \
            self.K_test, self.DE.dict_weights = for_lp_mission(name, self.graph, self.DE.nodes_list,
                                                               self.DE.dict_snapshots, self.DE.dict_weights, test_ratio,
                                                               self.start_index)
            
        self.graph_lp = self.DE.create_weighted_graph_for_all_times()

        print("Times in train are: ")
        print(list(self.DE.dict_snapshots.keys()))
        print("Number of edges in the train set: ", self.K_train, " | Number of edges in the test set: ", self.K_test)

        # test edges are deleted so we need to re-build it
        self.DE.g_list, self.DE.nodes_list, T, index = self.DE.create_cumulative_graphs(number)

        # calculate the embedding using FODGE
        self.full_dict_embeddings, self.dict_all_embeddings, self.total_time = self.DE.calculate_embedding()
        self.DE.full_dict_embeddings, self.DE.dict_all_embeddings = self.full_dict_embeddings, self.dict_all_embeddings
        self.DE.save_embedding(save_path, mission="lp1")

        t = time() - initial_t
        
        print(f"FODGE is done after {t} seconds. Starting temporal link prediction task")

        # perform temporal link prediction task
        self.params_dict = {"rounds": 1, "number_choose": 1}
        self.dict_initial = lp_mission(self.full_dict_embeddings, self.dict_snapshots, self.nodes, self.index,
                                       [test_ratio], self.params_dict["rounds"], self.params_dict["number_choose"],
                                       self.true_edges_train, self.true_edges_test, self.K_train, self.K_test,
                                       non_edges_file, self.start_index, self.DE.dim)
    

def create_keras_model(d):
    """
    Create the keras model of the task, d is the embedding dimension
    """
    model = Sequential()
    
    model.add(Dense(d, activation="relu", input_shape=(2*d, )))
    model.add(Dense(1))
    model.add(Activation('sigmoid'))
    
    opti = keras.optimizers.Adam(learning_rate=0.01)

    model.compile(optimizer="Adam", loss="binary_crossentropy", metrics=[tf.keras.metrics.AUC()])
    model.summary()
    
    return model


def fit_and_predict(model, X_train, X_test, y):
    """
    Fitting of the model and predicting
    :param model: Keras model
    :param X_train: Training samples
    :param X_test: Test samples
    :param y: Labels
    :return: Predicted values
    """
    y_train = y.reshape((-1,1))
    model.fit(X_train, y_train, epochs=5)
    print("done fitting")
    y_pred = model.predict(X_test)
    print("done predicting")
    return y_pred

    
def evaluate_edge_classification(X_train, X_test, Y_train, Y_test, d):
    """
    Predictions of nodes' labels.
    :param X_train: The features' graph- norm (for train set)
    :param X_test: The features' graph- norm (for tset set)
    :param Y_train: The edges labels- 0 for true, 1 for false (for train set)
    :param Y_test: The edges labels- 0 for true, 1 for false (for test set)
    :return: Scores- F1-macro, F1-micro accuracy and auc
    """
    model = create_keras_model(d)
    my_prediction = fit_and_predict(model, X_train, X_test, Y_train)
    prediction = my_prediction.ravel()
    fpr_keras, tpr_keras, thresholds_keras = roc_curve(Y_test, prediction)
    auc_keras = auc(fpr_keras, tpr_keras)
    # print("auc score is ", auc_keras)
    score = model.evaluate(X_test, Y_test, verbose=1)
    print("Test Loss: ", score[0])
    print("Test AUC: ", score[1])

    for i in range(my_prediction.size):
        a = my_prediction[i][0]
        if a > 0.5:
            my_prediction[i][0] = 1
        else:
            my_prediction[i][0] = 0
    
    acc = accuracy_score(Y_test, my_prediction)
    micro = f1_score(Y_test, my_prediction, average='micro')
    macro = f1_score(Y_test, my_prediction, average='macro')
    return micro, macro, acc, score[1]
    

def exp_lp(X_train, X_test, Y_train, Y_test, test_ratio_arr, rounds, d):
    """
    The final node classification task as explained in our git.
    :param X_train: The features' graph- norm (for train set)
    :param X_test: The features' graph- norm (for tset set)
    :param Y_train: The edges labels- 0 for true, 1 for false (for train set)
    :param Y_test: The edges labels- 0 for true, 1 for false (for test set)
    :param test_ratio_arr: To determine how to split the data into train and test. This an array with multiple options
                           of how to split.
    :param rounds: How many times we're doing the mission. Scores will be the average.
    :param d: Embedding Dimension
    :return: Scores for all splits and all splits- F1-micro, F1-macro accuracy and auc
    """

    micro = [None] * rounds
    macro = [None] * rounds
    acc = [None] * rounds
    auc = [None] * rounds

    for round_id in range(rounds):
        micro_round = [None] * len(test_ratio_arr)
        macro_round = [None] * len(test_ratio_arr)
        acc_round = [None] * len(test_ratio_arr)
        auc_round = [None] * len(test_ratio_arr)

        for i, test_ratio in enumerate(test_ratio_arr):
            micro_round[i], macro_round[i], acc_round[i], auc_round[i] = evaluate_edge_classification\
                (X_train, X_test, Y_train, Y_test, d)

        micro[round_id] = micro_round
        macro[round_id] = macro_round
        acc[round_id] = acc_round
        auc[round_id] = auc_round

    micro = np.asarray(micro)
    macro = np.asarray(macro)
    acc = np.asarray(acc)
    auc = np.asarray(auc)

    return micro, macro, acc, auc


def lp_mission(full_dict_proj, dict_snapshots, nodes, index, ratio_arr, rounds, number_choose, true_edges_train,
               true_edges_test, K_train, K_test, non_edges_file, start_index, d):
    """
    Dynamic Link prediction Task where one wants the scores as a function of size of the initial embedding.
    Notice test ratio must be fixed. The variable that changes here is the size of the initial embedding. For more
    explanation, see our pdf file attached in out git.
    :param full_dict_proj: Dict embeddings for all nodes that appear in the graph, no matter in which time stamps.
                           If a node appear in more than one time stamp, its final embedding is its emedding it the
                           last time it has appeared.
    :param dict_snapshots: Dict where keys==times and values==list of edges for each time stamp.
    :param nodes: List of nodes that appear both in test and train
    :param index: Index of pivot time- until pivot time (including) it is train set, afterwards it is test set.
    :param ratio_arr: Wanted test ratio (r = K_test / (K_train + K_test)
    :param rounds: How many times we're doing the mission. Scores will be the average.
    :param number_choose: Number of times to choose false edges randomly.
    :param true_edges_train: List of true edges for train set.
    :param true_edges_test: List of true edges for test set.
    :param K_train: Number of true edges in the train set.
    :param K_test: Number of true edges in the test set.
    :param non_edges_file: Csv file containing non edges in each time. First column is the time stamp, second is
                               source node and third is target node.
    :param start_index: Index of the cumulative initial graph
    :param d: Embedding dimension
    :return: Scores of link prediction task for each dataset- Micro-F1, Macro-F1, Accuracy and AUC. They return as
            lists for each size of initial embedding for each method
    """
    dict_initial = {}
    for r in ratio_arr:
        all_micro = []
        all_macro = []
        all_acc = []
        all_auc = []
        for j in range(1):
            my_micro, my_macro, my_acc, my_auc = initialize_scores()
            missing_edges_train, missing_edges_test = create_full_lists_missing_edges(index, non_edges_file, nodes, start_index, dict_snapshots, full_dict_proj)
            for i in range(number_choose):
                false_edges_train, false_edges_test = create_data_for_link_prediction(K_train, K_test, missing_edges_train, missing_edges_test)
                print(K_test, " : ", K_train)
                X_train, Y_train, dict_edges_train = create_x_y(full_dict_proj, true_edges_train, false_edges_train, K_train, d)
                X_test, Y_test, dict_edges_test = create_x_y(full_dict_proj, true_edges_test, false_edges_test, K_test, d)
                micro, macro, acc, auc = exp_lp(X_train, X_test, Y_train, Y_test, [r], rounds, d)
                avg_micro, avg_macro, avg_acc, avg_auc = calculate_all_avg_scores_lp(micro, macro, acc, auc, rounds)
                my_micro = first_help_calculate_lp(my_micro, avg_micro)
                my_macro = first_help_calculate_lp(my_macro, avg_macro)
                my_acc = first_help_calculate_lp(my_acc, avg_acc)
                my_auc = first_help_calculate_lp(my_auc, avg_auc)
            my_micro = second_help_calculate_lp(my_micro, number_choose)
            my_macro = second_help_calculate_lp(my_macro, number_choose)
            my_acc = second_help_calculate_lp(my_acc, number_choose)
            my_auc = second_help_calculate_lp(my_auc, number_choose)
            all_micro.append(my_micro[0])
            all_macro.append(my_macro[0])
            all_acc.append(my_acc[0])
            all_auc.append(my_auc[0])
        dict_initial.update({r: [all_micro, all_macro, all_acc, all_auc]})
    return dict_initial

